####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

""" This module provides semiconductor models.
"""

####################################################################################################

import numpy as np

####################################################################################################

import PySpice.Physics.PhysicalConstants as Cst

####################################################################################################

class ShockleyDiode:

    """ This class provides an implementation of the Shockley Diode Model.
    """

    ##############################################

    def __init__(self,
                 Is=10e-12, # 10 pA
                 n=1,
                 degree=25, kelvin=None):

        self.Is = Is # reverse bias saturation current
        self.n = n # ideality factor or emission coefficient
        self.T = Cst.temperature(degree=degree, kelvin=kelvin)

    ##############################################

    @property
    def Vt(self):
        """ Thermal Voltage """
        return Cst.kT(kelvin=self.T) / Cst.q

    ##############################################

    def I(self, Vd):
        return self.Is*(np.exp(Vd/(self.n*self.Vt)) - 1)

    ##############################################

    def rd(self, Vd):
        """ Dynamic resistance defined by dVd/dI. """
        return self.n*self.Vt/self.I(Vd)



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2020 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = ['NGSPICE_SUPPORTED_VERSION']

from .SimulationType import LAST_VERSION as NGSPICE_SUPPORTED_VERSION



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import logging
import os
import sys

from PySpice.Tools.Path import parent_directory_of

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

def find_libraries():

    try:
        library_path = os.environ['PySpiceLibraryPath']
    except KeyError:
        # Fixme: only works for one level
        python_file = os.path.abspath(sys.argv[0])
        examples_root = parent_directory_of(python_file, step=2)
        # .../PySpice/examples/diode/__example_rst_factory__nlrrr2fh.py .../PySpice/examples
        library_path = os.path.join(examples_root, 'libraries')

    _module_logger.info('SPICE library path is {}'.format(library_path))

    return library_path



####################################################################################################
#
# Fix custom dunders
#
# grep -R __ PySpice | grep '__ =' | sed -e 's/.*: *//' | sed -e 's/ =.*//' | sort | uniq
#
####################################################################################################

# pytest unit-test OK

# class VoltageDivider(SubCircuitFactory):
#     __name__ = 'VoltageDivider'
#     _nodes = ('input', 'output_plus', 'output_minus')

#   File "PySpice/Spice/Netlist.py", line 1156, in __init__
#     super().__init__(self.__name__, *self._nodes, **kwargs)
# TypeError: __init__() argument after * must be an iterable, not NoneType


####################################################################################################

from pathlib import Path
import os

####################################################################################################

special = (
)

internals = (
    '__binary_operator_map__',
    '__classes__',
    '__declaration_order__',
    '__hash_map__',
    '__name_to_unit__',
    '__nodes__',
    '__operators__',
    '__optional_parameters__',
    '__parameters_from_args__',
    '__positional_parameters__',
    '__prefixed_unit_map__',
    '__prefixes__',
    '__spice_to_parameters__',
    '__unary_operator_map__',
    '__unit_map__',
    '__units__',
    '__value_ctor__',
    '__values_ctor__',
    '__variable_cls__',
)

constants = (
    '__MAX_COMMAND_LENGTH__',
    '__VALID_KWARGS__',
    '__alias__',
    '__analysis_name__',
    '__as_unit__',
    '__base_units__',
    '__default_unit__',
    '__git_tag__',
    '__is_si__',
    '__long_alias__',
    '__number_of_operands__',
    '__operator__',
    '__pins__',
    '__power__',
    '__precedence__',
    '__prefix__',
    '__quantity__',
    '__scale__', # unused ?
    '__si_unit__',
    '__spice_prefix__',
    '__spice_suffix__', # unused ?
    '__unit_name__',
    '__unit_suffix__',
)

####################################################################################################

def fix_file(path):
    print(path)

    with open(path, 'r') as fh:
        content = fh.read()

    for dunder in internals:
        fixed_name = dunder.replace('__', '_')
        fixed_name = fixed_name[:-1]
        content = content.replace(dunder, fixed_name)
    for dunder in constants:
        fixed_name = dunder.upper().replace('__', '')
        content = content.replace(dunder, fixed_name)

    os.rename(path, str(path) + '~~~')
    with open(path, 'w') as fh:
        fh.write(content)

####################################################################################################

for root_directory in (
        'PySpice',
        'unit-test',
):
    for root, directories, filenames in os.walk(root_directory):
        for filename in filenames:
            path = Path(root, filename)
            if path.suffix == '.py':
                fix_file(path)



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module provides an interface to run ngspice in server mode and get back the simulation
output.

When ngspice runs in server mode, it writes on the standard output an header and then the simulation
output in binary format.  At the end of the simulation, it writes on the standard error a line of
the form:

    .. code::

        @@@ \d+ \d+

where the second number is the number of points of the simulation.  Due to the iterative and
adaptive nature of a transient simulation, the number of points is only known at the end.

Any line starting with "Error" in the standard output indicates an error in the simulation process.
The line "run simulation(s) aborted" in the standard error indicates the simulation aborted.

Any line starting with *Warning* in the standard error indicates non critical error in the
simulation process.

"""

####################################################################################################

import logging
import os
import re
import subprocess

####################################################################################################

from .RawFile import RawFile

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class SpiceServer:

    """This class wraps the execution of ngspice in server mode and convert the output to a Python data
    structure.

    Example of usage::

      spice_server = SpiceServer(spice_command='/path/to/ngspice')
      raw_file = spice_server(spice_input)

    It returns a :obj:`PySpice.Spice.RawFile` instance.

    """

    _logger = _module_logger.getChild('SpiceServer')

    SPICE_COMMAND = 'ngspice'

    ##############################################

    def __init__(self, **kwargs):

        self._spice_command = kwargs.get('spice_command') or self.SPICE_COMMAND

    ##############################################

    def _decode_number_of_points(self, line):

        """Decode the number of points in the given line."""

        match = re.match(r'@@@ (\d+) (\d+)', line)
        if match is not None:
            return int(match.group(2))
        else:
            raise NameError("Cannot decode the number of points")

    ##############################################

    def _parse_stdout(self, stdout):

        """Parse stdout for errors."""

        # self._logger.debug(os.linesep + stdout)

        error_found = False
        # UnicodeDecodeError: 'utf-8' codec can't decode byte 0xc0 in position 870: invalid start byte
        # lines = stdout.decode('utf-8').splitlines()
        lines = stdout.splitlines()
        for line_index, line in enumerate(lines):
            if line.startswith(b'Error '):
                error_found = True
                self._logger.error(os.linesep + line.decode('utf-8') + os.linesep + lines[line_index+1].decode('utf-8'))
        if error_found:
            raise NameError("Errors was found by Spice")

    ##############################################

    def _parse_stderr(self, stderr):

        """Parse stderr for warnings and return the number of points."""

        self._logger.debug(os.linesep + stderr)

        stderr_lines = stderr.splitlines()
        number_of_points = None
        for line in stderr_lines:
            if line.startswith('Warning:'):
                self._logger.warning(line[len('Warning :'):])
            elif line == 'run simulation(s) aborted':
                raise NameError('Simulation aborted' + os.linesep + stderr)
            elif line.startswith('@@@'):
                number_of_points = self._decode_number_of_points(line)

        return number_of_points

    ##############################################

    def __call__(self, spice_input):

        """Run SPICE in server mode as a subprocess for the given input and return a
        :obj:`PySpice.RawFile.RawFile` instance.

        """

        self._logger.info("Start the spice subprocess")

        process = subprocess.Popen((self._spice_command, '-s'),
                                   stdin=subprocess.PIPE,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE)
        input_ = str(spice_input).encode('utf-8')
        stdout, stderr = process.communicate(input_)
        # stdout = stdout.decode('utf-8')
        stderr = stderr.decode('utf-8')

        self._parse_stdout(stdout)
        number_of_points = self._parse_stderr(stderr)
        if number_of_points is None:
            raise NameError('The number of points was not found in the standard error buffer,'
                            ' ngspice returned:' + os.linesep +
                            stderr)

        return RawFile(stdout, number_of_points)



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

""" This module provides an implementation for enumerate.

The enumerate factory :func:`EnumFactory` builds a enumerate from a list of names and assigns to
these constants a value from 0 to N-1, where N is the number of constants.  For example::

  enum = EnumFactory('Enum1', ('cst1', 'cst2'))

builds a enumerate with *cst1* set to 0 and *cst2* set to 1.

We can get a constant's value using an integer context like::

  int(enum.cst1)

and the constant's name using::

  repr(enum.cst1)

We can test constant equality using::

  enum1.cst == enum2.cst

or with something that understand the *int* protocol::

  enum1.cst == obj
  # equivalent to
  int(enum1.cst) == int(obj)

The number of constants could be retrieved with::

  len(enum)

The enumerate factory :func:`ExplicitEnumFactory` is a variant that permits to specify the values of
the constants::

  enum2 = ExplicitEnumFactory('Enum2', {'cst1':1, 'cst2':3})

We can test if a value is in the enumerate using::

  constant_value in enum2

"""

####################################################################################################

# __all__ = ['EnumFactory', 'ExplicitEnumFactory']

####################################################################################################

class ReadOnlyMetaClass(type):

    """ This meta class implements a class where attributes are read only. """

    ##############################################

    def __setattr__(self, name, value):

        raise NotImplementedError

####################################################################################################

class EnumMetaClass(ReadOnlyMetaClass):

    """ This meta class implements the :func:`len` protocol. """

    ##############################################

    def __len__(self):

        return self._size

    ##############################################

    def __getitem__(self, i):

        return self._index[i]

####################################################################################################

class ExplicitEnumMetaClass(ReadOnlyMetaClass):

    """ This meta class implements the operator ``in``. """

    ##############################################

    def __contains__(self, item):

        return item in self.constants

####################################################################################################

class EnumConstant:

    """ Define an Enum Constant """

    ##############################################

    def __init__(self, name, value):

        self._name = name
        self._value = value

    ##############################################

    def __eq__(self, other):

        return self._value == int(other)

    ##############################################

    def __int__(self):

        return self._value

    ##############################################

    def __hash__(self):

        return self._value

    ##############################################

    def __repr__(self):

        return self._name

####################################################################################################

def EnumFactory(enum_name, enum_tuple):

    """ Return an :class:`EnumMetaClass` instance, where *enum_name* is the class name and
    *enum_tuple* is an iterable of constant's names.
    """

    index = [EnumConstant(name, value) for value, name in enumerate(enum_tuple)]

    obj_dict = {}
    obj_dict['_size'] = len(enum_tuple)
    obj_dict['_index'] = index
    obj_dict.update({str(enum):enum for enum in index})

    return EnumMetaClass(enum_name, (), obj_dict)

####################################################################################################

def ExplicitEnumFactory(enum_name, enum_dict):

    """ Return an :class:`ExplicitEnumMetaClass` instance, where *enum_name* is the class name and
    *enum_dict* is a dict of constant's names and their values.
    """

    obj_dict = {}
    obj_dict['constants'] = list(enum_dict.values())
    for name, value in list(enum_dict.items()):
        obj_dict[name] = EnumConstant(name, value)

    return ExplicitEnumMetaClass(enum_name, (), obj_dict)



#r# ============================
#r#  Internal Device Parameters
#r# ============================

#r# This example show how to access internal device parameters.

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Probe.Plot import plot
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit, SubCircuit, SubCircuitFactory
from PySpice.Unit import *

####################################################################################################

class Level2(SubCircuitFactory):
    NAME = 'level2'
    NODES = ('d4', 'g4', 'v4')
    def __init__(self):
        super().__init__()
        self.M(1, 'd4', 'g4', 'v4', 'v4', model='NMOS', w=1e-5, l=3.5e-7)

class Level1(SubCircuitFactory):
    NAME = 'level1'
    NODES = ('d3', 'g3', 'v3')
    def __init__(self):
        super().__init__()
        self.X('mos2', 'level2', 'd3', 'g3', 'v3')
        self.subcircuit(Level2())

circuit = Circuit('Transistor output characteristics')
circuit.V('dd', 'd1', circuit.gnd, 2)
circuit.V('ss', 'vsss', circuit.gnd, 0)
circuit.V('sig', 'g1', 'vsss', 0)
circuit.X('mos1', 'level1', 'd1', 'g1', 'vsss')

if True:
    circuit.subcircuit(Level1())
else:
    subcircuit_level2 = SubCircuit('level2', 'd4', 'g4', 'v4')
    subcircuit_level2.M(1, 'd4', 'g4', 'v4', 'v4', model='NMOS', w=1e-5, l=3.5e-7)

    subcircuit_level1 = SubCircuit('level1', 'd3', 'g3', 'v3')
    subcircuit_level1.X('mos2', 'level2', 'd3', 'g3', 'v3')
    subcircuit_level1.subcircuit(subcircuit_level2)

    circuit.subcircuit(subcircuit_level1)

circuit.model('NMOS', 'NMOS', LEVEL=8)

print(str(circuit))

####################################################################################################

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
# Fixme: python return code is not 0 on Windows if the following line is executed
#        but any error is reported
## analysis = simulator.dc(Vdd=slice(0, 5, .1)) # Fixme: ,Vsig=slice(1, 5, 1)

#r# To be completed ...

####################################################################################################

#?# vdd d1 0 2.0
#?# vss vsss 0 0
#?# vsig g1 vsss 0
#?# xmos1 d1 g1 vsss level1
#?# .subckt level1 d3 g3 v3
#?# xmos2 d3 g3 v3 level2
#?# .ends
#?# .subckt level2 d4 g4 v4
#?# m1 d4 g4 v4 v4 nmos w=1e-5 l=3.5e -007
#?# .ends
#?# .dc vdd 0 5 0.1 vsig 0 5 1
#?# .control
#?# save all @m.xmos1.xmos2.m1[vdsat]
#?# run
#?# plot vss#branch $ current measured at the top level
#?# plot @m.xmos1.xmos2.m1[vdsat]
#?# .endc
#?# .MODEL NMOS NMOS LEVEL=8
#?# +VERSION=3.2.4 TNOM=27 TOX=7.4E-9
#?# .end



#r# This example shows how to simulate and plot the characteristic curve of a Zener diode.

####################################################################################################

import numpy as np
import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Spice.Netlist import Circuit
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

#f# circuit_macros('zener-diode-characteristic-curve-circuit.m4')

circuit = Circuit('Diode DC Curve')

circuit.include(spice_library['1N4148'])
# 1N5919B: 5.6 V, 3.0 W Zener Diode Voltage Regulator
circuit.include(spice_library['d1n5919brl'])

circuit.V('input', 'in', circuit.gnd, 10@u_V)
circuit.R(1, 'in', 'out', 1@u_Ω) # not required for simulation
# circuit.X('D1', '1N4148', 'out', circuit.gnd)
circuit.X('DZ1', 'd1n5919brl', 'out', circuit.gnd)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.dc(Vinput=slice(-10, 2, .05)) # 10mV

figure, (ax1, ax2) = plt.subplots(2, figsize=(20, 10))

zener_part = analysis.out <= -5.4@u_V
# compute derivate
# fit linear part

ax1.grid()
# Fixme: scale
ax1.plot(analysis.out, -analysis.Vinput*1000)
ax1.axvline(x=0, color='black')
ax1.axvline(x=-5.6, color='red')
ax1.axvline(x=1, color='red')
ax1.legend(('Diode curve',), loc=(.1,.8))
ax1.set_xlabel('Voltage [V]')
ax1.set_ylabel('Current [mA]')

ax2.grid()
# Fixme:
# U = RI   R = U/I
dynamic_resistance = np.diff(-analysis.out) / np.diff(analysis.Vinput)
# ax2.plot(analysis.out[:-1], dynamic_resistance/1000)
ax2.semilogy(analysis.out[10:-1], dynamic_resistance[10:], base=10)
ax2.axvline(x=0, color='black')
ax2.axvline(x=-5.6, color='red')
ax2.legend(('Dynamic Resistance',), loc=(.1,.8))
ax2.set_xlabel('Voltage [V]')
ax2.set_ylabel('Dynamic Resistance [Ohm]')

# coefficients = np.polyfit(analysis.out[zener_part], dynamic_resistance[zener_part], deg=1)
# x = np.array((min(analysis.out[zener_part]), max(analysis.out[zener_part])))
# y = coefficients[0]*x + coefficients[1]
# axe.semilogy(x, y, 'red')

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'zener-characteristic-curve.png')



#r# This example shows the computation of the DC bias and sensitivity in a voltage divider.

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

#f# circuit_macros('voltage-divider.m4')

circuit = Circuit('Voltage Divider')

circuit.V('input', 'in', circuit.gnd, 10@u_V)
circuit.R(1, 'in', 'out', 9@u_kΩ)
circuit.R(2, 'out', circuit.gnd, 1@u_kΩ)

####################################################################################################

simulator = circuit.simulator(temperature=25, nominal_temperature=25)

analysis = simulator.operating_point()
for node in (analysis['in'], analysis.out): # .in is invalid !
    print('Node {}: {} V'.format(str(node), float(node)))
#o#

# Fixme: Xyce sensitivity analysis
analysis = simulator.dc_sensitivity('v(out)')
for element in analysis.elements.values():
    print(element, float(element))
#o#



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import unittest

####################################################################################################

from PySpice.Spice.BasicElement import *
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

class TestBasicElement(unittest.TestCase):

    ##############################################

    def _test_spice_declaration(self, element, spice_declaration):

        self.assertEqual(str(element), spice_declaration)

    ##############################################

    def test(self):

        self._test_spice_declaration(Resistor(Circuit(''), '1', 'n1', 'n2', 100),
                                     'R1 n1 n2 100')
        self._test_spice_declaration(Resistor(Circuit(''), '1', 'n1', 'n2', kilo(1)),
                                     'R1 n1 n2 1k')
        self._test_spice_declaration(Resistor(Circuit(''), '1', 'n1', 'n2', kilo(1),
                                              ac=kilo(2),
                                              multiplier=2,
                                              scale=1.5,
                                              temperature=25, device_temperature=26,
                                              noisy=True),
                                     'R1 n1 n2 1k ac=2k dtemp=26 m=2 noisy=1 scale=1.5 temp=25')
        self._test_spice_declaration(Resistor(Circuit(''), '1', 'n1', 'n2', kilo(1),
                                              noisy=False),
                                     'R1 n1 n2 1k')

        self._test_spice_declaration(XSpiceElement(Circuit(''), '1', 1, 0, model='cap'),
                                     'A1 1 0 cap')

####################################################################################################

if __name__ == '__main__':

    unittest.main()



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import os

from ..RawFile import VariableAbc, RawFileAbc

####################################################################################################

"""This module provide tools to read the output of Xyce.

Header

"""

####################################################################################################

import logging

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class Variable(VariableAbc):

    ##############################################

    def is_voltage_node(self):

        name = self.name.lower()
        return name.startswith('v(') or not self.is_branch_current()

    ##############################################

    def is_branch_current(self):
        return self.name.endswith('#branch')

    ##############################################

    @staticmethod
    def to_voltage_name(node):
        return 'v({})'.format(node)

    ##############################################

    @property
    def simplified_name(self):

        name = self.name
        if len(name) > 1 and name[1] == '(':
            return name[2:-1]
        elif name.endswith('#branch'):
            return name[:-7]
        elif '#' in name:
            # Xyce change name of type "output_plus" to "OUTPUT#PLUS"
            return name.replace('#', '_')
        else:
            return self.name

####################################################################################################

class RawFile(RawFileAbc):

    """ This class parse the stdout of ngspice and the raw data output.

    Public Attributes:

      :attr:`data`

      :attr:`date`

      :attr:`flags`
        'real' or 'complex'

      :attr:`number_of_points`

      :attr:`number_of_variables`

      :attr:`plot_name`
        AC Analysis, Operating Point, Sensitivity Analysis, DC transfer characteristic

      :attr:`title`

      :attr:`variables`

    """

    _logger = _module_logger.getChild('RawFile')

    _variable_cls = Variable

    ##############################################

    def __init__(self, output):

        raw_data = self._read_header(output)
        self._read_variable_data(raw_data)
        # self._to_analysis()

        self._simulation = None


    ##############################################

    def _read_header(self, output):

        """ Parse the header """

        # see https://github.com/FabriceSalvaire/PySpice/issues/132
        #   Xyce open the file in binary mode and print using: os << "Binary:" << std::endl;
        #   endl is thus \n
        binary_line = b'Binary:\n'
        binary_location = output.find(binary_line)
        if binary_location < 0:
            raise NameError('Cannot locate binary data')
        raw_data_start = binary_location + len(binary_line)
        self._logger.debug(os.linesep + output[:raw_data_start].decode('utf-8'))
        header_lines = output[:binary_location].splitlines()
        raw_data = output[raw_data_start:]
        header_line_iterator = iter(header_lines)

        self.title = self._read_header_field_line(header_line_iterator, 'Title')
        self.date = self._read_header_field_line(header_line_iterator, 'Date')
        self.plot_name = self._read_header_field_line(header_line_iterator, 'Plotname')
        self.flags = self._read_header_field_line(header_line_iterator, 'Flags')
        self.number_of_variables = int(self._read_header_field_line(header_line_iterator, 'No. Variables'))
        self.number_of_points = int(self._read_header_field_line(header_line_iterator, 'No. Points'))
        self._read_header_field_line(header_line_iterator, 'Variables')
        self._read_header_variables(header_line_iterator)

        return raw_data

    ##############################################

    def fix_case(self):

        """ Ngspice return lower case names. This method fixes the case of the variable names. """

        circuit = self.circuit
        element_translation = {element.upper():element for element in circuit.element_names}
        node_translation = {node.upper():node for node in circuit.node_names}
        for variable in self.variables.values():
            variable.fix_case(element_translation, node_translation)

    ##############################################

    def _to_dc_analysis(self):

        if 'sweep' in self.variables:
            sweep_variable = self.variables['sweep']
        else:
            raise NotImplementedError

        return super()._to_dc_analysis(sweep_variable)



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2020 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""Tool to perform post installation.

"""

####################################################################################################

from pathlib import Path
from zipfile import ZipFile
import argparse
import os
import shutil
import sys
import tempfile

import requests

####################################################################################################

# Archive:  resources/ngspice-31_dll_64.zip
#   Spice64_dll/dll-mingw/
#   Spice64_dll/dll-mingw/libgcc_s_seh-1.dll
#   Spice64_dll/dll-mingw/libgomp-1.dll
#   Spice64_dll/dll-mingw/libwinpthread-1.dll
#   Spice64_dll/dll-mingw/msys-ngspice-0.dll
#   Spice64_dll/dll-vs/
#   Spice64_dll/dll-vs/ngspice.dll
#   Spice64_dll/dll-vs/vcomp140.dll
#   Spice64_dll/include/
#   Spice64_dll/include/ngspice/
#   Spice64_dll/include/ngspice/sharedspice.h
#   Spice64_dll/lib/
#   Spice64_dll/lib/about-libs.txt
#   Spice64_dll/lib/lib-mingw/
#   Spice64_dll/lib/lib-mingw/libngspice.dll.a
#   Spice64_dll/lib/lib-mingw/libngspice.la
#   Spice64_dll/lib/lib-vs/
#   Spice64_dll/lib/lib-vs/ngspice.exp
#   Spice64_dll/lib/lib-vs/ngspice.lib
#   Spice64_dll/lib/ngspice/
#   Spice64_dll/lib/ngspice/analog.cm
#   Spice64_dll/lib/ngspice/digital.cm
#   Spice64_dll/lib/ngspice/spice2poly.cm
#   Spice64_dll/lib/ngspice/table.cm
#   Spice64_dll/lib/ngspice/xtradev.cm
#   Spice64_dll/lib/ngspice/xtraevt.cm
#   Spice64_dll/share/
#   Spice64_dll/share/ngspice/
#   Spice64_dll/share/ngspice/scripts/
#   Spice64_dll/share/ngspice/scripts/spinit

####################################################################################################

class CircuitTest:

    ##############################################

    def test_spinit(self):

        from PySpice.Spice.Netlist import Circuit
        import PySpice.Unit as U

        circuit = Circuit('Test')

        # Fixme: On Windows
        #   Supplies reduced to   2.5749% Supplies reduced to   1.7100% Warning: source stepping failed
        #   doAnalyses: Too many iterations without convergence
        source = circuit.V('cc', 'vcc', circuit.gnd, 15@U.u_V)
        circuit.R(1, 'output', 'comparator', 1@U.u_kΩ)
        circuit.C(1, 'comparator', circuit.gnd, 100@U.u_nF)
        circuit.R(2, 'output', 'reference', 100@U.u_kΩ)
        circuit.R(3, 'vcc', 'reference', 100@U.u_kΩ)
        circuit.R(4, 'reference', circuit.gnd, 100@U.u_kΩ)
        # circuit.NonLinearVoltageSource(1, 'output', circuit.gnd,
        #                                expression='V(reference, comparator)',
        #                                table=((-U.micro(1), 0),
        #                                       (U.micro(1), source.dc_value))
        # )

        simulator = circuit.simulator(temperature=25, nominal_temperature=25)
        simulator.initial_condition(comparator=0)  # Fixme: simulator.nodes.comparator == 0
        analysis = simulator.transient(step_time=1@U.u_us, end_time=500@U.u_us)

        if (len(analysis.output)) < 500:
            raise NameError('Simualtion failed')

####################################################################################################

class PySpicePostInstallation:

    GITHUB_URL = 'https://github.com/FabriceSalvaire/PySpice'

    NGSPICE_BASE_URL = 'https://sourceforge.net/projects/ngspice/files'
    NGSPICE_RELEASE_URL = NGSPICE_BASE_URL + '/ng-spice-rework'
    NGSPICE_WINDOWS_DLL_URL = NGSPICE_RELEASE_URL + '/{0}/ngspice-{0}_dll_64.zip'
    NGSPICE_WINDOWS_DLL_OLD_URL = NGSPICE_RELEASE_URL + '/old-releases/{0}/ngspice-{0}_dll_64.zip'
    NGSPICE_MANUAL_URL = NGSPICE_RELEASE_URL + '/{0}/ngspice-{0}-manual.pdf/download'
    NGSPICE_MANUAL_OLD_URL = NGSPICE_RELEASE_URL + '/old-releases/{0}/ngspice-{0}-manual.pdf/download'

    ##############################################

    def run(self):

        parser = argparse.ArgumentParser(description='Tool to perform PySpice Post Installation.')

        parser.add_argument(
            '--ngspice-version',
            type=int, default=None,
            help='NgSpice version to install',
        )

        parser.add_argument(
            '--install-ngspice-dll',
            action='store_true',
            help='install Windows DLL',
        )

        parser.add_argument(
            '--force-install-ngspice-dll',
            action='store_true',
            help='force DLL installation (for debug only)',
        )

        parser.add_argument(
            '--download-ngspice-manual',
            action='store_true',
            help='download Ngspice manual',
        )

        parser.add_argument(
            '--check-install',
            action='store_true',
            help='check installation',
        )

        parser.add_argument(
            '--download-example',
            action='store_true',
            help='download examples',
        )

        self._args = parser.parse_args()

        count = 0
        if self._args.install_ngspice_dll or self._args.force_install_ngspice_dll:
            self.install_ngspice_dll()
            count += 1

        if self._args.check_install:
            self.check_installation()
            count += 1

        if self._args.download_example:
            self.download_example()
            count += 1

        if self._args.download_ngspice_manual:
            self.download_ngspice_manual()
            count += 1

        if not count:
            parser.print_help()

    ##############################################

    def _download_file(self, url, dst_path):
        print('Get {} ... -> {}'.format(url, dst_path))
        response = requests.get(url, allow_redirects=True)
        if response.status_code != requests.codes.ok:
            response.raise_for_status()
        with open(dst_path, mode='wb') as fh:
            fh.write(response.content)

    ##############################################

    @property
    def ngspice_version(self):
        if not hasattr(self, '_ngspice_version'):
            version = self._args.ngspice_version
            if version is None:
                from PySpice.Spice.NgSpice import NGSPICE_SUPPORTED_VERSION
                version = NGSPICE_SUPPORTED_VERSION
            self._ngspice_version = version
        return self._ngspice_version

    ##############################################

    def install_ngspice_dll(self):

        if not(os.name == 'nt' or self._args.force_install_ngspice_dll):
            return

        from PySpice.Spice import NgSpice

        with tempfile.TemporaryDirectory() as tmp_directory:
            tmp_directory = Path(tmp_directory)
            url = self.NGSPICE_WINDOWS_DLL_URL.format(self.ngspice_version)
            zip_path = tmp_directory.joinpath('ngspice-{}_dll_64.zip'.format(self.ngspice_version))
            dst_path = Path(NgSpice.__file__).parent
            try:
                self._download_file(url, zip_path)
            except requests.exceptions.HTTPError:
                print('Download failed, trying another URL...')
                url = self.NGSPICE_WINDOWS_DLL_OLD_URL.format(self.ngspice_version)
                self._download_file(url, zip_path)
            with ZipFile(zip_path) as zip_file:
                zip_file.extractall(path=dst_path)
                print('Extracted {} in {}'.format(zip_path, dst_path.joinpath('Spice64_dll')))

        spice64_path = dst_path.joinpath('Spice64_dll')
        dll_path = spice64_path.joinpath('dll-vs')
        # src = dll_path.joinpath('ngspice-{}.dll'.format(self.ngspice_version))
        src = 'ngspice-{}.dll'.format(self.ngspice_version)
        target = dll_path.joinpath('ngspice.dll')
        # For ngspice version <=31 DLL naming did not contain a version number
        if dll_path.joinpath(src).exists():
            if target.exists():
                target.unlink()
            try:
                target.symlink_to(src)
            except OSError:
                # OSError: symbolic link privilege not held
                # Windows: If User Account Control (UAC) is on, any user with the "Create Symbolic
                #   Links" privilege that is not in the Administrators group can simply create a
                #   symbolic link.  For users within the Administrators group and with UAC on, the user
                #   must "Run as Administrator".
                shutil.copy(target.parent.joinpath(src), target)

        spinit_path = spice64_path.joinpath('share', 'ngspice', 'scripts', 'spinit')
        with open(spinit_path) as fh:
            content = fh.read()
        rule = '='*80
        print(rule)
        print(content)
        print(rule)
        cm_path = spice64_path.joinpath('lib', 'ngspice')
        content = content.replace('../lib/ngspice/', str(cm_path) + '/')
        print(rule)
        print(content)
        print(rule)
        with open(spinit_path, 'w') as fh:
            fh.write(content)

    ##############################################

    def download_ngspice_manual(self):
        url = self.NGSPICE_MANUAL_URL.format(self.ngspice_version)
        try:
            self._download_file(url, 'ngspice-manual-{}.pdf'.format(self.ngspice_version))
        except requests.exceptions.HTTPError:
            print('Download failed, trying another URL...')
            url = self.NGSPICE_MANUAL_OLD_URL.format(self.ngspice_version)
            self._download_file(url, 'ngspice-manual-{}.pdf'.format(self.ngspice_version))

    ##############################################

    def check_installation(self):

        """Tool to check PySpice is correctly installed.

        """

        import ctypes.util

        print('OS:', sys.platform)
        print()

        print('Environments:')
        for _ in (
                'PATH',
                'LD_LIBRARY_PATH',
                'PYTHONPATH',

                'NGSPICE_LIBRARY_PATH',
                'SPICE_LIB_DIR',
                'SPICE_EXEC_DIR',
                'SPICE_ASCIIRAWFILE',
                'SPICE_SCRIPTS',
                'NGSPICE_MEAS_PRECISION',
                'SPICE_NO_DATASEG_CHECK',
                'NGSPICE_INPUT_DIR',
        ):
            print(_, os.environ.get(_, 'undefined'))
        print()

        if 'VIRTUAL_ENV' in os.environ:
            print('On Virtual Environment:')
            for _ in (
                    'VIRTUAL_ENV',
            ):
                print(_, os.environ.get(_, 'undefined'))
            print()

        if 'CONDA_PREFIX' in os.environ:
            print('On Anaconda:')
            for _ in (
                    # not specific
                    'CONDA_EXE',
                    'CONDA_PYTHON_EXE',
                    # 'CONDA_SHLVL', # shell level, 1 in conda else 0
                    # '_CE_CONDA', # empty
                    # specific
                    'CONDA_DEFAULT_ENV',
                    'CONDA_PREFIX',
                    # 'CONDA_PROMPT_MODIFIER',
            ):
                print(_, os.environ.get(_, 'undefined'))
            print()

        try:
            print('Load PySpice module')
            import PySpice
            print('loaded {} version {}'.format(PySpice.__file__, PySpice.__version__))
            print()
        except ModuleNotFoundError:
            print('PySpice module not found')
            return

        import PySpice.Logging.Logging as Logging
        logger = Logging.setup_logging(logging_level='INFO')

        from PySpice.Config import ConfigInstall
        from PySpice.Spice import NgSpice
        from PySpice.Spice.NgSpice import NGSPICE_SUPPORTED_VERSION
        from PySpice.Spice.NgSpice.Shared import NgSpiceShared

        print('ngspice supported version:', NGSPICE_SUPPORTED_VERSION)
        print()

        ##############################################

        message = os.linesep.join((
            'NgSpiceShared configuration is',
            '  NgSpiceShared.NGSPICE_PATH = {0.NGSPICE_PATH}',
            '  NgSpiceShared.LIBRARY_PATH = {0.LIBRARY_PATH}',
        ))
        print(message.format(NgSpiceShared))
        print()

        ##############################################

        cwd = Path(os.curdir).resolve()
        print('Working directory:', cwd)
        print()

        locale_ngspice = cwd.joinpath('ngspice-{}'.format(NGSPICE_SUPPORTED_VERSION))
        if locale_ngspice.exists() and locale_ngspice.is_dir():
            print('Found local ngspice:')
            for root, _, filenames in os.walk(locale_ngspice, followlinks=True):
                for filename in filenames:
                    print(root, filename)
            print()


        ngspice_module_path = Path(NgSpice.__file__).parent
        print('NgSpice:', ngspice_module_path)
        for root, _, filenames in os.walk(ngspice_module_path):
            for filename in filenames:
                print(root, filename)
        print()

        ##############################################

        if ConfigInstall.OS.on_windows:
            print('OS is Windows')
            library = NgSpiceShared.LIBRARY_PATH
        elif ConfigInstall.OS.on_osx:
            print('OS is OSX')
            library = 'ngspice'
        elif ConfigInstall.OS.on_linux:
            print('OS is Linux')
            library = 'ngspice'
        else:
            raise NotImplementedError

        library_path = ctypes.util.find_library(library)
        print('Found in library search path: {}'.format(library_path))

        ##############################################

        print('\nLoad NgSpiceShared')
        ngspice = NgSpiceShared.new_instance(verbose=True)

        if ConfigInstall.OS.on_linux:
            # For Linux see DLOPEN(3)
            # Apparently there is no simple way to get the path of the loaded library ...
            # But we can look in the process maps
            pid = os.getpid()
            maps_path = '/proc/{}/maps'.format(pid)
            with open(maps_path) as fh:
                for line in fh:
                    if '.so' in line and 'ngspice' in line:
                        parts = [x for x in line.split() if x]
                        path = parts[-1]
                        print('loaded {}'.format(path))
                        break
        print()

        if ngspice.spinit_not_found:
            print('WARNING: spinit was not found')
            print()

        message = os.linesep.join((
            'Ngspice version is {0.ngspice_version}',
            '  has xspice: {0.has_xspice}',
            '  has cider {0.has_cider}',
        ))
        print(message.format(ngspice))
        print()

        command = 'version -f'
        print('> ' + command)
        print(ngspice.exec_command(command))
        print()

        circuit_test = CircuitTest()
        circuit_test.test_spinit()

        print('PySpice should work as expected')

    ##############################################

    def download_example(self):

        import PySpice
        version = PySpice.__version__

        RELEASE_URL = self.GITHUB_URL + '/archive/v{}.zip'.format(version)

        zip_path = 'examples.zip'

        dst_path = input("Enter the path where you want to extract examples: ")
        dst_path = Path(dst_path).resolve()
        dst_parent = dst_path.parent
        if not dst_parent.exists():
            print("Directory {} doesn't exists".format(dst_parent))
            return

        with tempfile.TemporaryDirectory() as tmp_directory:
            self._download_file(RELEASE_URL, zip_path)
            with ZipFile(zip_path) as zip_file:
                zip_file.extractall(path=tmp_directory)
            examples_path = Path(tmp_directory).joinpath('PySpice-{}'.format(version), 'examples')
            shutil.copytree(examples_path, dst_path)
            print('Extracted examples in {}'.format(dst_path))

####################################################################################################

def main():
    _ = PySpicePostInstallation()
    return _.run()



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module implements SPICE circuit elements.

.. warning:: Some elements are partially implemented.

.. warning:: The API documentation generated by Sphinx is perfectible for this module. The source
             code can be more informative.

.. note:: It would be nice to have a useful and working documentation in the interactive environment.

The element's parameters are internally implemented using class attributes and descriptors.

There are two types of parameters, positionals which are required and where the order mater, and
optional parameters which are passed as keyword arguments.

Parameters are registered with more expressive and longer names than their Spice counterparts.  For
optional parameters, we can use both Spice and longer name for convenience.

See Ngspice documentation for details.

+--------------+------------------------------------------------------+
| First letter + Element description                                  |
+--------------+------------------------------------------------------+
| A            + XSPICE code model                                    |
+--------------+------------------------------------------------------+
| B            + Behavioral (arbitrary) source                        |
+--------------+------------------------------------------------------+
| C            + Capacitor                                            |
+--------------+------------------------------------------------------+
| D            + Diode                                                |
+--------------+------------------------------------------------------+
| E            + Voltage-controlled voltage source (VCVS)             |
+--------------+------------------------------------------------------+
| F            + Current-controlled current source (CCCs)             |
+--------------+------------------------------------------------------+
| G            + Voltage-controlled current source (VCCS)             |
+--------------+------------------------------------------------------+
| H            + Current-controlled voltage source (CCVS)             |
+--------------+------------------------------------------------------+
| I            + Current source                                       |
+--------------+------------------------------------------------------+
| J            + Junction field effect transistor (JFET)              |
+--------------+------------------------------------------------------+
| K            + Coupled (Mutual) Inductors                           |
+--------------+------------------------------------------------------+
| L            + Inductor                                             |
+--------------+------------------------------------------------------+
| M            + Metal oxide field effect transistor (MOSFET)         |
+--------------+------------------------------------------------------+
| N            + Numerical device for GSS                             |
+--------------+------------------------------------------------------+
| O            + Lossy transmission line                              |
+--------------+------------------------------------------------------+
| P            + Coupled multiconductor line (CPL)                    |
+--------------+------------------------------------------------------+
| Q            + Bipolar junction transistor (BJT)                    |
+--------------+------------------------------------------------------+
| R            + Resistor                                             |
+--------------+------------------------------------------------------+
| S            + Switch (voltage-controlled)                          |
+--------------+------------------------------------------------------+
| T            + Lossless transmission line                           |
+--------------+------------------------------------------------------+
| U            + Uniformly distributed RC line                        |
+--------------+------------------------------------------------------+
| V            + Voltage source                                       |
+--------------+------------------------------------------------------+
| W            + Switch (current-controlled)                          |
+--------------+------------------------------------------------------+
| X            + Subcircuit                                           |
+--------------+------------------------------------------------------+
| Y            + Single lossy transmission line (TXL)                 |
+--------------+------------------------------------------------------+
| Z            + Metal semiconductor field effect transistor (MESFET) |
+--------------+------------------------------------------------------+

"""

####################################################################################################

import logging

from ..Tools.StringTools import str_spice, join_list, join_dict
from ..Unit import U_m, U_s, U_A, U_V, U_Degree, U_Ω, U_F, U_H, U_Hz
from .Netlist import (Element, AnyPinElement, FixedPinElement, NPinElement, OptionalPin)
from .ElementParameter import (
    # KeyValueParameter,
    BoolKeyParameter,
    ElementNamePositionalParameter,
    ExpressionKeyParameter,
    ExpressionPositionalParameter,
    FlagParameter,
    FloatKeyParameter,
    FloatPairKeyParameter,
    FloatTripletKeyParameter,
    FloatPositionalParameter,
    InitialStatePositionalParameter,
    IntKeyParameter,
    ModelPositionalParameter,
    )

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class DipoleElement(FixedPinElement):
    """This class implements a base class for dipole element."""
    PINS = ('plus', 'minus')

class TwoPortElement(FixedPinElement):
    """This class implements a base class for two-port element."""
    PINS = ('output_plus', 'output_minus', 'input_plus', 'input_minus')

####################################################################################################

class SubCircuitElement(NPinElement):

    """This class implements a sub-circuit.

    Spice syntax:

    .. code-block:: none

        XYYYYYY node1 node2 ... subcircuit_name parameter1=value1 ...

    Attributes:

      :attr:`subcircuit_name`

    .. note:: As opposite to Spice, the circuit's name is specified before the nodes so as to act as `*args`.

    """

    ALIAS = 'X'
    PREFIX = 'X'

    subcircuit_name = ElementNamePositionalParameter(position=0, key_parameter=False)

    ##############################################

    def __init__(self, netlist, name, subcircuit_name, *nodes, **parameters):

        super().__init__(netlist, name, nodes, subcircuit_name)

        # Fixme: match parameters to subcircuit
        self.parameters = parameters

        # Fixme: investigate
        # for key, value in parameters.items():
        #     parameter = KeyValueParameter(key)
        #     parameter.__set__(self, value)
        #     self.optional_parameters[key] = parameter
        #     setattr(self, key, parameter)

    ##############################################

    def copy_to(self, netlist):

        element = self.__class__(netlist, self._name, self.subcircuit_name, *self.node_names, **self.parameters)
        # Element.copy_to(self, element)
        return element

    ##############################################

    def format_spice_parameters(self):
        """ Return the formatted list of parameters. """

        spice_parameters = super().format_spice_parameters()
        if self.parameters:
            spice_parameters += ' ' + join_dict(self.parameters)

        return spice_parameters

####################################################################################################
#
# Elementary devices: Resistor, Capacitor, Inductor, Switch (VCSw/CCSw)
#
####################################################################################################

class Resistor(DipoleElement):

    """This class implements a resistor.

    Spice syntax:

    .. code-block:: none

        RXXXXXXX n+ n- value <ac=val> <m=val> <scale=val> <temp=val> <dtemp=val> <noisy=0|1>

    Keyword Parameters:

      :attr:`ac`

      :attr:`multiplier`
         alias `m`

      :attr:`scale`

      :attr:`temperature`
         alias `temp`

      :attr:`device_temperature`
         alias `dtemp`

      :attr:`noisy`

    Attributes:

      :attr:`resistance`

      :attr:`ac`

      :attr:`multiplier`

      :attr:`scale`

      :attr:`temperature`

      :attr:`device_temperature`

      :attr:`noisy`

    """

    ALIAS = 'R'
    PREFIX = 'R'

    resistance = FloatPositionalParameter(position=0, key_parameter=False, unit=U_Ω)
    ac = FloatKeyParameter('ac', unit=U_Ω)
    multiplier = IntKeyParameter('m')
    scale = FloatKeyParameter('scale')
    temperature = FloatKeyParameter('temp', unit=U_Degree)
    device_temperature = FloatKeyParameter('dtemp', unit=U_Degree)
    noisy = BoolKeyParameter('noisy')

####################################################################################################

class SemiconductorResistor(DipoleElement):

    """This class implements a Semiconductor resistor.

    Spice syntax:

    .. code-block:: none

        RXXXXXXX n+ n- <value> <mname> <l=length> <w=width> <temp=val> <dtemp=val> m=<val> <ac=val> <scale=val> <noisy=0|1>

    Keyword Parameters:

      :attr:`model`

      :attr:`length`
         alias `l`

      :attr:`width`
         alias `w`

      :attr:`temperature`
         alias `temp`

      :attr:`device_temperature`
         alias `dtemp`

      :attr:`multiplier`
         alias `m`

      :attr:`ac`

      :attr:`scale`

      :attr:`noisy`

    Attributes:

      :attr:`resistance`

      :attr:`model`

      :attr:`length`

      :attr:`width`

      :attr:`temperature`

      :attr:`device_temperature`

      :attr:`multiplier`

      :attr:`ac`

      :attr:`scale`

      :attr:`noisy`

    """

    ALIAS = 'SemiconductorResistor'
    PREFIX = 'R'

    resistance = FloatPositionalParameter(position=0, key_parameter=False, unit=U_Ω)
    model = ModelPositionalParameter(position=1, key_parameter=True)
    length = FloatKeyParameter('l', unit=U_m)
    width = FloatKeyParameter('w', unit=U_m)
    temperature = FloatKeyParameter('temp', unit=U_Degree)
    device_temperature = FloatKeyParameter('dtemp', unit=U_Degree)
    multiplier = IntKeyParameter('m')
    ac = FloatKeyParameter('ac', unit=U_Ω)
    scale = FloatKeyParameter('scale')
    noisy = BoolKeyParameter('noisy')

####################################################################################################

class BehavioralResistor(DipoleElement):

    # Behavioral / Behavioural

    """This class implements a behavioral resistor.

    Spice syntax:

    .. code-block:: none

        RXXXXXXX n+ n- 'expression' <tc1=value> <tc2=value>
        Rxxxxxxx n+ n- R='expression' <tc1=value> <tc2=value>

    Keyword Parameters:

      :attr:`tc1`

      :attr:`tc2`

    Attributes:

      :attr:`resistance_expression`

      :attr:`tc1`

      :attr:`tc2`

    """

    ALIAS = 'BehavioralResistor'
    PREFIX = 'R'

    resistance_expression = ExpressionPositionalParameter(position=0, key_parameter=False)
    tc1 = FloatKeyParameter('tc1')
    tc2 = FloatKeyParameter('tc2')

####################################################################################################

class Capacitor(DipoleElement):

    """This class implements a capacitor.

    Spice syntax:

    .. code-block:: none

        CXXXXXXX n+ n- <value> <mname> <m=val> <scale=val> <temp=val> <dtemp=val> <ic=init_condition>

    Keyword Parameters:

      :attr:`model`

      :attr:`multiplier`
         alias `m`

      :attr:`scale`

      :attr:`temperature`
         alias `temp`

      :attr:`device_temperature`
         alias `dtemp`

      :attr:`initial_condition`
         alias `ic`

    Attributes:

      :attr:`capacitance`

      :attr:`model`

      :attr:`multiplier`

      :attr:`scale`

      :attr:`temperature`

      :attr:`device_temperature`

      :attr:`initial_condition`

    """

    ALIAS = 'C'
    PREFIX = 'C'

    capacitance = FloatPositionalParameter(position=0, key_parameter=False, unit=U_F)
    model = ModelPositionalParameter(position=1, key_parameter=True)
    multiplier = IntKeyParameter('m')
    scale = FloatKeyParameter('scale')
    temperature = FloatKeyParameter('temp', unit=U_Degree)
    device_temperature = FloatKeyParameter('dtemp', unit=U_Degree)
    initial_condition = FloatKeyParameter('ic')

####################################################################################################

class SemiconductorCapacitor(DipoleElement):

    """This class implements a semiconductor capacitor.

    Spice syntax:

    .. code-block:: none

        CXXXXXXX n+ n- <value> <mname> <l=length> <w=width> <m=val> <scale=val> <temp=val> <dtemp=val> <ic=init_condition>

    Keyword Parameters:

      :attr:`model`

      :attr:`length`
         alias `l`

      :attr:`width`
         alias `w`

      :attr:`multiplier`
         alias `m`

      :attr:`scale`

      :attr:`temperature`
         alias `temp`

      :attr:`device_temperature`
         alias `dtemp`

      :attr:`initial_condition`
         alias `ic`

    Attributes:

      :attr:`capacitance`

      :attr:`model`

      :attr:`length`

      :attr:`width`

      :attr:`multiplier`

      :attr:`scale`

      :attr:`temperature`

      :attr:`device_temperature`

      :attr:`initial_condition`

    """

    ALIAS = 'SemiconductorCapacitor'
    PREFIX = 'C'

    capacitance = FloatPositionalParameter(position=0, key_parameter=False, unit=U_F)
    model = ModelPositionalParameter(position=1, key_parameter=True)
    length = FloatKeyParameter('l', unit=U_m)
    width = FloatKeyParameter('w', unit=U_m)
    multiplier = IntKeyParameter('m')
    scale = FloatKeyParameter('scale')
    temperature = FloatKeyParameter('temp', unit=U_Degree)
    device_temperature = FloatKeyParameter('dtemp', unit=U_Degree)
    initial_condition = FloatKeyParameter('ic')

####################################################################################################

class BehavioralCapacitor(DipoleElement):

    """This class implements a behavioral capacitor.

    Spice syntax:

    .. code-block:: none

        CXXXXXXX n+ n- 'expression' <tc1=value> <tc2=value>
        CXXXXXXX n+ n- C='expression' <tc1=value> <tc2=value>

    Keyword Parameters:

      :attr:`tc1`

      :attr:`tc2`

    Attributes:

      :attr:`capacitance_expression`

      :attr:`tc1`

      :attr:`tc2`

    """

    ALIAS = 'BehavioralCapacitor'
    PREFIX = 'C'

    capacitance_expression = ExpressionPositionalParameter(position=0, key_parameter=False)
    tc1 = FloatKeyParameter('tc1')
    tc2 = FloatKeyParameter('tc2')

####################################################################################################

class Inductor(DipoleElement):

    """This class implements an inductor.

    Spice syntax:

    .. code-block:: none

        LYYYYYYY n+ n- <value> <mname> <nt=val> <m=val> <scale=val> <temp=val> <dtemp=val> <ic=init_condition>

    Keyword Parameters:

      :attr:`nt`

      :attr:`multiplier`
         alias `m`

      :attr:`scale`

      :attr:`temperature`
         alias `temp`

      :attr:`device_temperature`
         alias `dtemp`

      :attr:`initial_condition`
         alias `ic`

    Attributes:

      :attr:`inductance`

      :attr:`model`

      :attr:`nt`

      :attr:`multiplier`

      :attr:`scale`

      :attr:`temperature`

      :attr:`device_temperature`

      :attr:`initial_condition`

    """

    ALIAS = 'L'
    PREFIX = 'L'

    inductance = FloatPositionalParameter(position=0, key_parameter=False, unit=U_H)
    model = ModelPositionalParameter(position=1, key_parameter=True)
    nt = FloatKeyParameter('nt')
    multiplier = IntKeyParameter('m')
    scale = FloatKeyParameter('scale')
    temperature = FloatKeyParameter('temp', unit=U_Degree)
    device_temperature = FloatKeyParameter('dtemp', unit=U_Degree)
    initial_condition = FloatKeyParameter('ic')

####################################################################################################

class BehavioralInductor(DipoleElement):

    """This class implements a behavioral inductor.

    Spice syntax:

    .. code-block:: none

        LXXXXXXX n+ n- 'expression' <tc1=value> <tc2=value>
        LXXXXXXX n+ n- L='expression' <tc1=value> <tc2=value>

    Keyword Parameters:

      :attr:`tc1`

      :attr:`tc2`

    Attributes:

      :attr:`inductance_expression`

      :attr:`tc1`

      :attr:`tc2`

    """

    ALIAS = 'BehavioralInductor'
    PREFIX = 'L'

    inductance_expression = ExpressionPositionalParameter(position=0, key_parameter=False)
    tc1 = FloatKeyParameter('tc1')
    tc2 = FloatKeyParameter('tc2')

####################################################################################################

class CoupledInductor(AnyPinElement):

    """This class implementss a coupled (mutual) inductors.

    Spice syntax:

    .. code-block:: none

        KXXXXXXX LYYYYYYY LZZZZZZZ value

    Keyword Parameters:

    Attributes:

      :attr:`inductor1`

      :attr:`inductor2`

      :attr:`coupling_factor`

    """

    ALIAS = 'K'
    PREFIX = 'K'

    inductor1 = ElementNamePositionalParameter(position=0, key_parameter=False)
    inductor2 = ElementNamePositionalParameter(position=1, key_parameter=False)
    coupling_factor = FloatPositionalParameter(position=2, key_parameter=False)

    _logger = _module_logger.getChild('CoupledInductor')

 ##############################################

    def __init__(self, name, *args, **kwargs):

        super().__init__(name, *args, **kwargs)

        self._inductors = []
        for inductor in (self.inductor1, self.inductor2):
            try:
                self.netlist.element(inductor)
            except KeyError:
                try:
                    inductor = 'L' + inductor
                    self.netlist.element(inductor)
                    self._logger.info('Prefixed element {}'.format(inductor))
                except KeyError:
                    raise ValueError('Element with name {} not found'.format(inductor))
            # Fixme: str or Element instance ?
            self._inductors.append(inductor)
        self.inductor1, self.inductor2 = self._inductors

####################################################################################################

class VoltageControlledSwitch(TwoPortElement):

    """This class implements a voltage controlled switch.

    Spice syntax:

    .. code-block:: none

        SXXXXXXX n+ n- nc+ nc- model <on> <off>

    Keyword Parameters:

      :attr:`model`

      :attr:`initial_state`

    Attributes:

      :attr:`model`

      :attr:`initial_state`

    """

    ALIAS = 'S'
    LONG_ALIAS = 'VCS'
    PREFIX = 'S'

    model = ModelPositionalParameter(position=0, key_parameter=True)
    initial_state = InitialStatePositionalParameter(position=1, key_parameter=True)

####################################################################################################

class CurrentControlledSwitch(DipoleElement):

    """This class implements a current controlled switch.

    Spice syntax:

    .. code-block:: none

        WYYYYYYY n+ n- vname model <on> <off>

    Keyword Parameters:

      :attr:`source`

      :attr:`model`

      :attr:`initial_state`

    Attributes:

      :attr:`source`

      :attr:`model`

      :attr:`initial_state`

    """

    ALIAS = 'W'
    LONG_ALIAS = 'CCS'
    PREFIX = 'W'

    source = ElementNamePositionalParameter(position=0, key_parameter=True)
    model = ModelPositionalParameter(position=1, key_parameter=True)
    initial_state = InitialStatePositionalParameter(position=2, key_parameter=True)

####################################################################################################
#
# Voltage and Current Sources
#
####################################################################################################

class VoltageSource(DipoleElement):

    """This class implements an independent sources for voltage.

    Spice syntax:

    .. code-block:: none

        VXXXXXXX n+ n- <<dc> dc/tran value> <ac <acmag <acphase>>> <distof1 <f1mag <f1phase>>> <distof2 <f2mag <f2phase>>>

    Keyword Parameters:

    Attributes:

      :attr:`dc_value`

    """

    ALIAS = 'V'
    PREFIX = 'V'

    # Fixme: ngspice manual doesn't describe well the syntax
    dc_value = FloatPositionalParameter(position=0, key_parameter=False, unit=U_V)

####################################################################################################

class CurrentSource(DipoleElement):

    """This class implements an independent sources for current.

    Spice syntax:

    .. code-block:: none

       IYYYYYYY n+ n- <<dc> dc/tran value> <ac <acmag <acphase>>> <distof1 <f1mag <f1phase>>> <distof2 <f2mag <f2phase>>>

    Keyword Parameters:

    Attributes:

      :attr:`dc_value`

    """

    ALIAS = 'I'
    PREFIX = 'I'

    # Fixme: ngspice manual doesn't describe well the syntax
    dc_value = FloatPositionalParameter(position=0, key_parameter=False, unit=U_A)

####################################################################################################

class VoltageControlledCurrentSource(TwoPortElement):

    """This class implements a linear voltage-controlled current sources (VCCS).

    Spice syntax:

    .. code-block:: none

        Gxxx n+ n- nc+ nc- value <m=val>

    Keyword Parameters:

      :attr:`multiplier`
         alias `m`

    Attributes:

      :attr:`transconductance`

    """

    ALIAS = 'VCCS'
    PREFIX = 'G'

    transconductance = ExpressionPositionalParameter(position=0, key_parameter=False)
    multiplier = IntKeyParameter('m')

####################################################################################################

class VoltageControlledVoltageSource(TwoPortElement):

    """This class implements a linear voltage-controlled voltage sources (VCVS).

    Spice syntax:

    .. code-block:: none

        EXXXXXXX n+ n- nc+ nc- value

    Keyword Parameters:

    Attributes:

      :attr:`voltage_gain`

    """

    ALIAS = 'VCVS'
    PREFIX = 'E'

    voltage_gain = ExpressionPositionalParameter(position=0, key_parameter=False)

####################################################################################################

class CurrentControlledCurrentSource(DipoleElement):

    """This class implements a linear current-controlled current sources (CCCS).

    Spice syntax:

    .. code-block:: none

       FXXXXXXX n+ n- vname value <m=val>

    Keyword Parameters:

      :attr:`multiplier`
         alias `m`

    Attributes:

      :attr:`source`

      :attr:`current_gain`

    """

    ALIAS = 'F'
    LONG_ALIAS = 'CCCS'
    PREFIX = 'F'

    source = ElementNamePositionalParameter(position=0, key_parameter=False)
    current_gain = ExpressionPositionalParameter(position=1, key_parameter=False)
    multiplier = IntKeyParameter('m')

####################################################################################################

class CurrentControlledVoltageSource(DipoleElement):

    """This class implements a linear current-controlled voltage sources (CCVS).

    Spice syntax:

    .. code-block:: none

        HXXXXXXX n+ n- vname value

    Keyword Parameters:

    Attributes:

      :attr:`source`

      :attr:`transresistance`

    """

    ALIAS = 'H'
    LONG_ALIAS = 'CCVS'
    PREFIX = 'H'

    source = ElementNamePositionalParameter(position=0, key_parameter=False)
    transresistance = ExpressionPositionalParameter(position=1, key_parameter=False)

####################################################################################################
#
# Non-Linear Dependent Sources (Behavioral Sources)
#
####################################################################################################

class BehavioralSource(DipoleElement):

    """This class implements a behavioral source.

    Spice syntax:

    .. code-block:: none

        BXXXXXXX n+ n- <i=expr> <v=expr> <tc1=value> <tc2=value> <temp=value> <dtemp=value>

    Keyword Parameters:

      :attr:`current_expression`
          alias `i`

      :attr:`voltage_expression`
          alias `v`

      :attr:`tc1`

      :attr:`tc2`

      :attr:`temperature`
         alias `temp`

      :attr:`device_temperature`
         alias `dtemp`

    Attributes:

      :attr:`current_expression`

      :attr:`voltage_expression`

      :attr:`tc1`

      :attr:`tc2`

      :attr:`temperature`

      :attr:`device_temperature`

    """

    ALIAS = 'B'
    PREFIX = 'B'

    current_expression = ExpressionKeyParameter('i')
    voltage_expression = ExpressionKeyParameter('v')
    tc1 = FloatKeyParameter('tc1')
    tc2 = FloatKeyParameter('tc2')
    temperature = FloatKeyParameter('temp', unit=U_Degree)
    device_temperature = FloatKeyParameter('dtemp', unit=U_Degree)

####################################################################################################

class NonLinearVoltageSource(DipoleElement):

    """This class implements a non-linear voltage source.

    .. warning:: Partially implemented

    Spice syntax:

    .. code-block:: none

        Exxx n+ n- vol='expr'
        Exxx n+ n- value={expr}
        Exxx n1 n2 TABLE {expression}=(x0,y0) (x1,y1) (x2,y2)
        Exxx n+ n- ( POLY (nd) ) nc1+ nc1- ( nc2+ nc2- ... ) p0 ( p1 ... )
        Laplace

    Keyword Parameters:

    Attributes:

    """

    ALIAS = 'NonLinearVoltageSource'
    PREFIX = 'E'

    # Fixme:
    VALID_KWARGS = ('expression', 'table')

    ##############################################

    def __init__(self, name, *args, **kwargs):

        super().__init__(name, *args, **kwargs)

        self.expression = kwargs.get('expression', None)
        self.table = kwargs.get('table', None)

    ##############################################

    def __str__(self):

        spice_element = self.format_node_names()
        # Fixme: expression
        if self.table is not None:
            # TABLE {expression} = (x0, y0) (x1, y1) ...
            table = ['({}, {})'.format(str_spice(x), str_spice(y)) for x, y in self.table]
            spice_element += ' TABLE {%s} = %s' % (self.expression, join_list(table))
        return spice_element

####################################################################################################

class NonLinearCurrentSource(DipoleElement):

    """This class implements a non-linear current sources.

    .. warning:: Partially implemented

    Spice syntax:

    .. code-block:: none

        Gxxx n+ n- value={expr}
        Gxxx n1 n2 TABLE {expression}=(x0,y0) (x1,y1) (x2,y2)
        Gxxx n+ n- ( POLY (nd) ) nc1+ nc1- ( nc2+ nc2- ... ) p0 ( p1 ... )
        Laplace

    Keyword Parameters:

    Attributes:

      :attr:`transconductance`

    """

    ALIAS = 'NonLinearCurrentSource'
    PREFIX = 'G'

    transconductance = ExpressionPositionalParameter(position=0, key_parameter=False)

####################################################################################################
#
# Diode
#
####################################################################################################

class Diode(FixedPinElement):

    """This class implements a junction diode.

    Spice syntax:

    .. code-block:: none

        DXXXXXXX n+ n- mname <area=val> <m=val> <pj=val> <off> <ic=vd> <temp=val> <dtemp=val>

    Keyword Parameters:

      :attr:`model`

      :attr:`area`

      :attr:`multiplier`
         alias `m`

      :attr:`pj`

      :attr:`off`

      :attr:`ic`

      :attr:`temperature`
         alias `temp`

      :attr:`device_temperature`
         alias `dtemp`

    Attributes:

      :attr:`model`

      :attr:`area`

      :attr:`multiplier`

      :attr:`pj`

      :attr:`off`

      :attr:`ic`

      :attr:`temperature`

      :attr:`device_temperature`

    """

    ALIAS = 'D'
    PREFIX = 'D'
    PINS = (('cathode', 'plus'), ('anode', 'minus'))

    model = ModelPositionalParameter(position=0, key_parameter=True)
    area = FloatKeyParameter('area')
    multiplier = IntKeyParameter('m')
    pj = FloatKeyParameter('pj')
    off = FlagParameter('off')
    ic = FloatPairKeyParameter('ic')
    temperature = FloatKeyParameter('temp', unit=U_Degree)
    device_temperature = FloatKeyParameter('dtemp', unit=U_Degree)

####################################################################################################
#
# BJTs
#
####################################################################################################

class BipolarJunctionTransistor(FixedPinElement):

    """This class implements a bipolar junction transistor.

    Spice syntax:

    .. code-block:: none

         QXXXXXXX nc nb ne <ns> mname <area=val> <areac=val> <areab=val> <m=val> <off> <ic=vbe,vce> <temp=val> <dtemp=val>

    Keyword Parameters:

      :attr:`model`

      :attr:`area`

      :attr:`areac`

      :attr:`areab`

      :attr:`multiplier`
         alias `m`

      :attr:`off`

      :attr:`ic`

      :attr:`temperature`
         alias `temp`

      :attr:`device_temperature`
         alias `dtemp`

    Attributes:

      :attr:`model`

      :attr:`area`

      :attr:`areac`

      :attr:`areab`

      :attr:`multiplier`

      :attr:`off`

      :attr:`ic`

      :attr:`temperature`

      :attr:`device_temperature`

    """

    # Fixme: off doesn't fit in kwargs !

    ALIAS = 'Q'
    LONG_ALIAS = 'BJT'
    PREFIX = 'Q'
    PINS = ('collector', 'base', 'emitter', OptionalPin('substrate'))

    model = ModelPositionalParameter(position=0, key_parameter=True)
    area = FloatKeyParameter('area')
    areac = FloatKeyParameter('areac')
    areab = FloatKeyParameter('areab')
    multiplier = IntKeyParameter('m')
    off = FlagParameter('off')
    ic = FloatPairKeyParameter('ic')
    temperature = FloatKeyParameter('temp', unit=U_Degree)
    device_temperature = FloatKeyParameter('dtemp', unit=U_Degree)

####################################################################################################
#
# JFETs
#
####################################################################################################

class JfetElement(FixedPinElement):
    PINS = ('drain', 'gate', 'source')

class JunctionFieldEffectTransistor(JfetElement):

    """This class implements a bipolar junction transistor.

    Spice syntax:

    .. code-block:: none

         JXXXXXXX nd ng ns mname <area> <off> <ic=vds,vgs> <temp=t>

    Keyword Parameters:

      :attr:`model`

      :attr:`area`

      :attr:`off`

      :attr:`ic`

      :attr:`temperature`
         alias `temp`

    Attributes:

      :attr:`model`

      :attr:`area`

      :attr:`off`

      :attr:`ic`

      :attr:`temperature`

    """

    # Fixme: off doesn't fit in kwargs !

    ALIAS = 'J'
    LONG_ALIAS = 'JFET'
    PREFIX = 'J'

    model = ModelPositionalParameter(position=0, key_parameter=True)
    area = FloatKeyParameter('area')
    multiplier = IntKeyParameter('m')
    off = FlagParameter('off')
    ic = FloatPairKeyParameter('ic')
    temperature = FloatKeyParameter('temp', unit=U_Degree)

####################################################################################################
#
# MESFETs
#
####################################################################################################

class Mesfet(JfetElement):

    """This class implements a Metal Semiconductor Field Effect Transistor.

    Spice syntax:

    .. code-block:: none

         ZXXXXXXX nd ng ns mname <area> <off> <ic=vds,vgs>

    Keyword Parameters:

      :attr:`model`

      :attr:`area`

      :attr:`off`

      :attr:`ic`

    Attributes:

      :attr:`model`

      :attr:`area`

      :attr:`off`

      :attr:`ic`

    """

    # Fixme: off doesn't fit in kwargs !

    ALIAS = 'Z'
    LONG_ALIAS = 'MESFET'
    PREFIX = 'Z'

    model = ModelPositionalParameter(position=0, key_parameter=True)
    area = FloatKeyParameter('area')
    multiplier = IntKeyParameter('m')
    off = FlagParameter('off')
    ic = FloatPairKeyParameter('ic')

####################################################################################################
#
# MOSFETs
#
####################################################################################################

class Mosfet(FixedPinElement):

    """This class implements a Metal Oxide Field Effect Transistor.

    Spice syntax:

    .. code-block:: none

         MXXXXXXX nd ng ns nb mname <m=val> <l=val> <w=val>
         + <ad=val> <as=val> <pd=val> <ps=val> <nrd=val>
         + <nrs=val> <off> <ic=vds,vgs,vbs> <temp=t>

    Keyword Parameters:

      :attr:`model`

      :attr:`multiplier`
         alias `m`

      :attr:`length`
         alias `l`

      :attr:`width`
         alias `w`

      :attr:`nfin`
        only for Xyce

      :attr:`drain_area`
          alias `ad`

      :attr:`source_area`
          alias `as`

      :attr:`drain_perimeter`
          alias `pd`

      :attr:`source_perimeter`
          alias `ps`

      :attr:`drain_number_square`
          alias `nrd`

      :attr:`source_number_square`
          alias `nrs`

      :attr:`off`

      :attr:`ic`

      :attr:`temperature`
         alias `temp`

    Attributes:

      :attr:`model`

      :attr:`multiplier`

      :attr:`length`

      :attr:`width`

      :attr:`nfin`
        only for Xyce

      :attr:`drain_area`

      :attr:`source_area`

      :attr:`drain_perimeter`

      :attr:`source_perimeter`

      :attr:`drain_number_square`

      :attr:`source_number_square`

      :attr:`off`

      :attr:`ic`

      :attr:`temperature`

    """

    # Fixme: off doesn't fit in kwargs !

    ALIAS = 'M'
    LONG_ALIAS = 'MOSFET'
    PREFIX = 'M'
    PINS = ('drain', 'gate', 'source', ('bulk', 'substrate'))

    model = ModelPositionalParameter(position=0, key_parameter=True)
    multiplier = IntKeyParameter('m')
    length = FloatKeyParameter('l', unit=U_m)
    width = FloatKeyParameter('w', unit=U_m)
    drain_area = FloatKeyParameter('ad')
    source_area = FloatKeyParameter('as')
    drain_perimeter = FloatKeyParameter('pd')
    source_perimeter = FloatKeyParameter('ps')
    drain_number_square = FloatKeyParameter('nrd')
    source_number_square = FloatKeyParameter('nrs')
    off = FlagParameter('off')
    ic = FloatTripletKeyParameter('ic')
    temperature = FloatKeyParameter('temp', unit=U_Degree)

    # only for Xyce
    nfin = IntKeyParameter('nfin')

####################################################################################################
#
# Transmission Lines
#
####################################################################################################

class LosslessTransmissionLine(TwoPortElement):

    """This class implements a lossless transmission line.

    Spice syntax:

    .. code-block:: none

        TXXXXXXX N1 N2 N3 N4 Z0=VALUE <TD=VALUE> <F=FREQ <NL=NRMLEN>> <IC=V1, I1, V2, I2>

    where TD or F, NL must be specified.

    Keyword Parameters:

      :attr:`impedance`
         alias:`Z0`
         is the characteristic impedance

      :attr:`time_delay`
         alias:`TD`
         is the transmission delay

      :attr:`frequency`
         alias:`F`

      :attr:`normalized_length`
         alias:`NL`

    Attributes:

      :attr:`impedance`

      :attr:`time_delay`

      :attr:`frequency`

      :attr:`normalized_length`

    The transmission delay, `td`, may be specified directly (as `td=10ns`, for example).
    Alternatively, a frequency `f` may be given, together with `nl`, the normalized electrical
    length of the transmission line with respect to the wavelength in the line at the frequency
    `f`. If a frequency is specified but `nl` is omitted, 0.25 is assumed (that is, the frequency is
    assumed to be the quarter-wave frequency). Note that although both forms for expressing the line
    length are indicated as optional, one of the two must be specified.

    Note: Either time_delay or frequency must be given.

    """

    ALIAS = 'TransmissionLine'
    PREFIX = 'T'

    impedance = FloatKeyParameter('Z0', default=50, unit=U_Ω)
    time_delay = FloatKeyParameter('TD', unit=U_s)
    frequency = FloatKeyParameter('F', unit=U_Hz)
    normalized_length = FloatKeyParameter('NL')

    ##############################################

    def __init__(self, name, *args, **kwargs):

        super().__init__(name, *args, **kwargs)

        if not (self.has_parameter('time_delay') or
                (self.has_parameter('frequency') and self.has_parameter('normalized_length'))):
            raise NameError('Either TD or F, NL must be specified')

####################################################################################################

class LossyTransmission(TwoPortElement):

    """This class implements lossy transmission lines.

    Spice syntax:

    .. code-block:: none

        OXXXXXXX n1 n2 n3 n4 model

    Attributes:

      :attr:`model`

    .. note:: As opposite to Spice, the model is specified before the nodes so as to act as `*args`.

    """

    ALIAS = 'O'
    PREFIX = 'O'

    model = ModelPositionalParameter(position=0, key_parameter=True)

####################################################################################################

class CoupledMulticonductorLine(NPinElement):

    """This class implements coupled multiconductor lines.

    Spice syntax:

    .. code-block:: none

        PXXXXXXX NI1 NI2 ... NIX GND1 NO1 NO2 ... NOX GND2 model <len=length>

    Attributes:

      :attr:`model`

      :attr:`length`
         alias `len`
         length of the line in meters

    .. note:: As opposite to Spice, the model is specified before the nodes so as to act as `*args`.

    """

    ALIAS = 'P'
    PREFIX = 'P'

    model = ModelPositionalParameter(position=0, key_parameter=True)
    length = FloatKeyParameter('len', unit=U_m)

    ##############################################

    def __init__(self, netlist, name, *nodes, **parameters):

        super().__init__(netlist, name, nodes, **parameters)

####################################################################################################

class UniformDistributedRCLine(FixedPinElement):

    """This class implements uniform distributed RC lines.

    Spice syntax:

    .. code-block:: none

        UXXXXXXX n1 n2 n3 model l=length <n=number_of_lumps>

    Attributes:

      :attr:`model`

      :attr:`length`
         alias `l`
         length of the RC line in meters

      :attr:`number_of_lumps`
         alias `n`

    .. note:: As opposite to Spice, the model is specified before the nodes so as to act as `*args`.

    """

    ALIAS = 'U'
    PREFIX = 'U'
    PINS = ('output', 'input', 'capacitance_node')

    model = ModelPositionalParameter(position=0, key_parameter=True)
    length = FloatKeyParameter('l', unit=U_m)
    number_of_lumps = IntKeyParameter('n')

####################################################################################################

class SingleLossyTransmissionLine(TwoPortElement):

    # Fixme: special TwoPortElement

    """This class implements single lossy transmission lines.

    Spice syntax:

    .. code-block:: none

        YXXXXXXX N1 0 N2 0 model <len=length>

    Attributes:

      :attr:`model`

      :attr:`length`
         alias `len`
         length of the line in meters

    .. note:: As opposite to Spice, the model is specified before the nodes so as to act as `*args`.

    """

    ALIAS = 'Y'
    PREFIX = 'Y'

    model = ModelPositionalParameter(position=0, key_parameter=True)
    length = FloatKeyParameter('len', unit=U_m)

####################################################################################################
#
# XSPICE
#
####################################################################################################

class XSpiceElement(NPinElement):

    """This class implements a sub-circuit.

    Spice syntax:

    .. code-block:: none

        AXXXXXXX <%v ,%i ,%vd ,%id ,%g,%gd ,%h,%hd , or %d>
        + <[> <~><%v ,%i ,%vd ,%id ,%g,%gd ,%h,%hd , or %d>
        + <NIN1 or +NIN1 -NIN1 or "null">
        + <~>...< NIN2 .. <]> >
        + <%v ,%i ,%vd ,%id ,%g,%gd ,%h,%hd ,%d or %vnam >
        + <[> <~><%v ,%i ,%vd ,%id ,%g,%gd ,%h,%hd ,
        or %d>< NOUT1 or +NOUT1 -NOUT1 >
        + <~>...< NOUT2 .. <]>>
        + MODELNAME

        . MODEL MODELNAME MODELTYPE
        + <( PARAMNAME1 = <[> VAL1 <VAL2 ... <]>> PARAMNAME2 ..>)>

    Attributes:

      :attr:`model`

    .. note:: As opposite to Spice, the model is specified before the nodes so as to act as `*args`.

    .. warning:: Partially implemented.
    """

    ALIAS = 'A'
    PREFIX = 'A'

    model = ModelPositionalParameter(position=0, key_parameter=True)

    ##############################################

    def __init__(self, netlist, name, *nodes, **parameters):

        # Fixme: ok ???

        super().__init__(netlist, name, nodes, **parameters)

####################################################################################################
#
# GSS
#
####################################################################################################

class GSSElement(NPinElement):

    """This class implements GSS device.

    .. warning:: Not implemented
    """

    ALIAS = 'N'
    PREFIX = 'N'

    ##############################################

    def __init__(self):

        raise NotImplementedError



#skip#

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.NgSpice.Shared import NgSpiceShared

####################################################################################################

class MyNgSpiceShared(NgSpiceShared):

    ##############################################

    def get_vsrc_data(self, voltage, time, node, ngspice_id):
        self._logger.debug('ngspice_id-{} get_vsrc_data @{} node {}'.format(ngspice_id, time, node))
        voltage[0] = 1.
        return 0

    ##############################################

    def get_isrc_data(self, current, time, node, ngspice_id):
        self._logger.debug('ngspice_id-{} get_isrc_data @{} node {}'.format(ngspice_id, time, node))
        current[0] = 1.
        return 0

####################################################################################################

time_step = '1m'

# circuit = """ .title rc circuit
# * V1 1 0 1
# V1 1 0 dc 0 external
# R1 1 2 1
# C1 2 0 1 ic=0
# .tran {} 3 uic
# .end
# """.format(time_step)

circuit = """
.title Low-Pass RC Filter
Vinput in 0 DC 0V AC SIN(0V 1V 50Hz 0s 0)
Rf in out 1k
Cf out 0 1u
.options TNOM = 25
.options TEMP = 25
.ic
.ac dec 10 1 1Meg
.end
"""

ngspice_shared = MyNgSpiceShared(send_data=False)
ngspice_shared.load_circuit(circuit)
ngspice_shared.run()
print(ngspice_shared.plot_names)
# vectors = ngspice_shared.vectors('tran1')
# print(vectors)
# vectors = ngspice_shared.vectors('ac1')
# print(vectors)
# print(vectors['frequency'])
# print(ngspice_shared.plot('ac1'))
analysis = ngspice_shared.plot('ac1').to_analysis()

print(analysis.nodes)
print(analysis.branches)



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

class Manufacturer:

    ##############################################

    def __init__(self,
                 name,
                 url=None,
    ):

        self._name = name
        self._url = url

    ##############################################

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

    @property
    def url(self):
        return self._url

    @url.setter
    def url(self, value):
        self._url = value

####################################################################################################

class Footprint:

    ##############################################

    def __init__(self,
                 name,
    ):

        self._name = name

    ##############################################

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value

####################################################################################################

class Device:

    ##############################################

    def __init__(self,
                 name,
                 manufacturer,
                 datasheet_url=None,
                 model_url=None
    ):

        # part
        # part_number
        # footprint
        # description
        # device_category x/y
        # pins
        # features / parameters

        self._name = name

    ##############################################

    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, value):
        self._name = value



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2020 jmgc / Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import pickle
import tempfile
import unittest

import numpy as np

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Probe.WaveForm import WaveForm
from PySpice.Unit.Unit import UnitValues
from PySpice.Unit import u_kHz

####################################################################################################

class TestPickle(unittest.TestCase):

    ##############################################

    def test_ndarray(self):
        array = np.ndarray((1, 1))
        with tempfile.TemporaryFile() as fh:
            pickle.dump(array, fh)
            fh.seek(0)
            new_array = pickle.load(fh)
        self.assertEqual(array, new_array)

    ##############################################

    def test_unit_values(self):
        unit_values = UnitValues(u_kHz(100).prefixed_unit, (1, 1))
        new_unit_values = pickle.loads(pickle.dumps(unit_values))
        self.assertEqual(unit_values, new_unit_values)

    ##############################################

    def test_waveformx(self):
        waveform = WaveForm('Test', u_kHz(100).prefixed_unit, (1, 1))
        new_waveform = pickle.loads(pickle.dumps(waveform))
        self.assertEqual(waveform, new_waveform)

####################################################################################################

if __name__ == '__main__':
    unittest.main()



#r# =======================
#r#  How to Use SubCircuit
#r# =======================

#r# This example shows how to use subcircuits.

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.Netlist import Circuit, SubCircuit, SubCircuitFactory
from PySpice.Unit import *

####################################################################################################

#r# There is two ways to define subcircuit with PySpice, either using
#r# :class:`PySpice.Spice.Netlist.SubCircuit` or a simpler alternative
#r# :class:`PySpice.Spice.Netlist.SubCircuitFactory`.

#r#
#r# Let define a parallel resistor subcircuit using the :class:`PySpice.Spice.Netlist.SubCircuitFactory`

class ParallelResistor(SubCircuitFactory):
    NAME = 'parallel_resistor'
    NODES = ('n1', 'n2')
    def __init__(self, R1=1@u_Ω, R2=2@u_Ω):
        super().__init__()
        self.R(1, 'n1', 'n2', R1)
        self.R(2, 'n1', 'n2', R2)

#r# Let define a circuit

circuit = Circuit('Test')

#r# then we can use this subcircuit like this

circuit.subcircuit(ParallelResistor(R2=3@u_Ω))
circuit.X('1', 'parallel_resistor', 1, circuit.gnd)

print(circuit)
#o#

#r# If the above way is not suited for your purpose we can use this second approach

class ParallelResistor2(SubCircuit):
    NODES = ('n1', 'n2')
    def __init__(self, name, R1=1@u_Ω, R2=2@u_Ω):
        SubCircuit.__init__(self, name, *self.NODES)
        self.R(1, 'n1', 'n2', R1)
        self.R(2, 'n1', 'n2', R2)

circuit = Circuit('Test')
circuit.subcircuit(ParallelResistor2('pr1', R2=2@u_Ω))
circuit.X('1', 'pr1', 1, circuit.gnd)
circuit.subcircuit(ParallelResistor2('pr2', R2=3@u_Ω))
circuit.X('2', 'pr2', 1, circuit.gnd)

print(circuit)
#o#



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module implements mathematical functions.
"""

####################################################################################################

import math

####################################################################################################

def odd(x):
    """Return True is *x* is odd"""
    return x & 1

def even(x):
    """Return True is *x* is even"""
    return not(odd(x))

####################################################################################################

def rms_to_amplitude(x):
    """Return :math:`x \sqrt{2}`"""
    return x * math.sqrt(2)

def amplitude_to_rms(x):
    """Return :math:`x / \sqrt{2}`"""
    return x / math.sqrt(2)



#r# =========================================
#r#  Pass Raw Spice Definitions to a Netlist
#r# =========================================

#Fixme: to be documented, improved

#r# This example shows how to pass raw spice definitions to a netlist.

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

#r# Let define a circuit.

circuit = Circuit('Test')

#r# Pass raw Spice definitions to a circuit, aka netlist, content is inserted at the beginning of
#r# the netlist.
circuit.raw_spice = '''
Vinput in 0 10V
R1 in out 9kOhm
'''

#r# Pass element parameters as raw Spice, content is concatenated with `R2 out 0`
circuit.R(2, 'out', 0, raw_spice='1k')

print(circuit)
#o#



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import argparse

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.Parser import SpiceParser

####################################################################################################

def main():

    parser = argparse.ArgumentParser(description='Convert a circuit file to PySpice')

    parser.add_argument('circuit_file', # metavar='circuit_file',
                        help='.cir file')

    parser.add_argument('-o', '--output',
                        default=None,
                        help='Output file')

    parser.add_argument('--ground',
                        type=int,
                        default=0,
                        help='Ground node')

    parser.add_argument('--build',
                        default=False, action='store_true',
                        help='Build circuit')

    args = parser.parse_args()

    ##############################################

    parser = SpiceParser(path=args.circuit_file)

    if args.build:
        parser.build_circuit()

    circuit = parser.to_python_code(ground=args.ground)
    if args.output is not None:
        with open(args.output, 'w') as f:
            f.write(circuit)
    else:
        print(circuit)



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import logging
import re

####################################################################################################

from ..Tools.File import Directory
from .Parser import SpiceParser

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

class SpiceLibrary:

    """This class implements a Spice sub-circuits and models library.

    A library is a directory which is recursively scanned for '.lib' file and parsed for sub-circuit
    and models definitions.

    Example of usage::

        spice_library = SpiceLibrary('/some/path/')

    If the directory hierarchy contains a file that define a 1N4148 sub-circuit then we can retrieve
    the file path using::

        spice_library['1N4148']

    """

    _logger = _module_logger.getChild('Library')

    EXTENSIONS = (
        '.spice',
        '.lib',
        '.mod',
        '.lib@xyce',
        '.mod@xyce',
    )

    ##############################################

    def __init__(self, root_path, recurse=False, section=None):

        self._directory = Directory(root_path).expand_vars_and_user()

        self._subcircuits = {}
        self._models = {}

        for path in self._directory.iter_file():
            extension = path.extension.lower()
            if extension in self.EXTENSIONS:
                self._logger.debug("Parse {}".format(path))
                try:
                    spice_parser = SpiceParser(path=path, recurse=recurse, section=section)
                    for lib in spice_parser.incl_libs:
                        self._subcircuits.update(lib._subcircuits)
                        self._models.update(lib._models)
                except Exception as e:
                    # Parse problem with this file, so skip it and keep going.
                    self._logger.warn("Problem parsing {path} - {e}".format(**locals()))
                    continue
                if spice_parser.is_only_subcircuit():
                    for subcircuit in spice_parser.subcircuits:
                        name = self._suffix_name(subcircuit.name, extension)
                        self._subcircuits[name] = path
                elif spice_parser.is_only_model():
                    for model in spice_parser.models:
                        name = self._suffix_name(model.name, extension)
                        self._models[name] = path

    ##############################################

    @staticmethod
    def _suffix_name(name, extension):

        if extension.endswith('@xyce'):
            name += '@xyce'

        return name

    ##############################################

    def __getitem__(self, name):

        if name in self._subcircuits:
            return self._subcircuits[name]
        elif name in self._models:
            return self._models[name]
        else:
            # print('Library {} not found in {}'.format(name, self._directory))
            # self._logger.warn('Library {} not found in {}'.format(name, self._directory))
            raise KeyError(name)

    ##############################################

    @property
    def subcircuits(self):
        """ Dictionary of sub-circuits """
        return iter(self._subcircuits)

    @property
    def models(self):
        """ Dictionary of models """
        return iter(self._models)

    # ##############################################

    # def iter_on_subcircuits(self):
    #     return self._subcircuits.itervalues()

    # ##############################################

    # def iter_on_models(self):
    #     return self._models.itervalues()

    # ##############################################

    def search(self, s):
        """ Return dict of all models/subcircuits with names matching regex s. """
        matches = {}
        models_subcircuits = {**self._models, **self._subcircuits}
        for name, mdl_subckt in models_subcircuits.items():
            if re.search(s, name):
                matches[name] = mdl_subckt
        return matches



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

__all__ = [
    'join_dict',
    'join_lines'
    'join_list',
    'str_spice',
    'str_spice_list',
]

####################################################################################################

import os

####################################################################################################

from PySpice.Unit.Unit import UnitValue

####################################################################################################

def str_spice(obj, unit=True):

    # Fixme: right place ???

    """Convert an object to a Spice compatible string."""

    if isinstance(obj, UnitValue):
        if unit:
            return obj.str_spice()
        else: # Fixme: ok ???
            return obj.str(spice=False, space=False, unit=False)
    else:
        return str(obj)

####################################################################################################

def str_spice_list(*args):
    return [str_spice(x) for x in args]

####################################################################################################

def join_lines(items, prefix=''):
    return os.linesep.join([prefix + str(item)
                            for item in items
                            if item is not None]) # Fixme: and item

####################################################################################################

def join_list(items):
    # return ' '.join([str_spice(item)
    #                  for item in items
    #                  if item is not None and str_spice(item)])
    values = []
    for item in items:
        if item is not None:
            str_value = str_spice(item)
            if str_value:
                values.append(str_value)
    return ' '.join(values)

####################################################################################################

def join_dict(d):
    return ' '.join(["{}={}".format(key, str_spice(value))
                     for key, value in sorted(d.items())
                     if value is not None])



#r# =======================
#r#  Netlist Manipulations
#r# =======================

#r# This example shows how to manipulate netlist.

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.Netlist import Circuit, SubCircuitFactory
from PySpice.Unit import *

####################################################################################################

class SubCircuit1(SubCircuitFactory):
    NAME = 'sub_circuit1'
    NODES = ('n1', 'n2')
    def __init__(self):
        super().__init__()
        self.R(1, 'n1', 'n2', 1@u_Ω)
        self.R(2, 'n1', 'n2', 2@u_Ω)

#r# Let define a circuit.

circuit = Circuit('Test')

#r# When we add an element to a circuit, we can get a reference to it or ignore it:

C1 = circuit.C(1, 0, 1, 1@u_uF)

circuit.C(2, 1, 2, 2@u_uF)
circuit.subcircuit(SubCircuit1())
circuit.X('1', 'sub_circuit1', 2, 0)

#r# We can get back an element of a circuit using its name, either as a class attribute or using the
#r# dictionary interface:

C1 = circuit.C1
C1 = circuit['C1']

#r# and modify it

C1.capacitance = 10@u_F

#r# To get the SPICE netlist of a citcuit, we just have to convert it to a string:

print(circuit) # str(circuit) is implicit here

#r# same apply to an element

print(C1)

#r# We can disable an element in the circuit

C1.enabled = False
print(circuit)

#r# We can clone a circuit to another one

circuit2 = circuit.clone(title='A clone') # title is optional
print(circuit2)

#r# We can remove an element

C2 = circuit2.C2.detach()
print(circuit2)



#r# This example illustrates the diode recovery time and the capacitive behaviour of a PN junction.

# Fixme: Split the plots ? Add some explanations at the end

####################################################################################################

import numpy as np
import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Probe.Plot import plot
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

#r# Let define some parameters

dc_offset = 1@u_V
ac_amplitude = 100@u_mV

####################################################################################################

#r# We will first compute some quiescent points and the corresponding dynamic resistance.

#f# circuit_macros('diode-characteristic-curve-circuit.m4')

#r# Since this circuit is equivalent to a voltage divider, we can write the following relation :
#r#
#r# .. math::
#r#
#r#     V_{out} = \frac{Z_d}{R_1 + Z_d} V_{in}
#r#
#r# where :math:`Z_d` is the diode impedance.

circuit = Circuit('Diode')
circuit.include(spice_library['BAV21'])
# Fixme: Xyce: Device model BAV21: Illegal parameter(s) given for level 1 diode: IKF
source = circuit.V('input', 'in', circuit.gnd, dc_offset)
circuit.R(1, 'in', 'out', 1@u_kΩ)
circuit.D('1', 'out', circuit.gnd, model='BAV21')

quiescent_points = []
for voltage in (dc_offset - ac_amplitude, dc_offset, dc_offset + ac_amplitude):
    source.dc_value = voltage
    simulator = circuit.simulator(temperature=25, nominal_temperature=25)
    analysis = simulator.operating_point()
    # Fixme: handle unit
    quiescent_voltage = float(analysis.out)
    quiescent_current = - float(analysis.Vinput)
    quiescent_points.append(dict(voltage=voltage,
                                 quiescent_voltage=quiescent_voltage,
                                 quiescent_current=quiescent_current))
    print("Quiescent Point {:.1f} mV {:.1f} mA".format(quiescent_voltage*1e3, quiescent_current*1e3))
#o#

dynamic_resistance = ((quiescent_points[ 0]['quiescent_voltage'] -
                       quiescent_points[-1]['quiescent_voltage'])
                      /
                      (quiescent_points[ 0]['quiescent_current'] -
                       quiescent_points[-1]['quiescent_current']))

#?# print("Dynamic Resistance = {:.1f} Ω".format(dynamic_resistance))
#?# #o#

#r# We found a dynamic resistance of @<@dynamic_resistance:.1f@>@ Ω.

####################################################################################################

#r#
#r# We will now drive the diode with a sinusoidal source and perform an AC analysis.

#f# circuit_macros('diode-characteristic-curve-circuit-ac.m4')

circuit = Circuit('Diode')
circuit.include(spice_library['BAV21'])
circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd,
                   dc_offset=dc_offset, offset=dc_offset,
                   amplitude=ac_amplitude)
R = circuit.R(1, 'in', 'out', 1@u_kΩ)
circuit.D('1', 'out', circuit.gnd, model='BAV21')

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.ac(start_frequency=10@u_kHz, stop_frequency=1@u_GHz, number_of_points=10,  variation='dec')

#r# Let plot the voltage across the diode and the dynamic resistance as a function of the frequency.

figure, (ax1, ax2, ax3)  = plt.subplots(ncols=3, figsize=(20, 10))

# Fixme: handle unit in plot (scale and legend)
ax1.semilogx(analysis.frequency, np.absolute(analysis.out)*1e3)
ax1.grid(True)
ax1.grid(True, which='minor')
ax1.set_xlabel("Frequency [Hz]")
ax1.set_ylabel("Vd [mV]")

current = (analysis['in'] - analysis.out) / float(R.resistance)
ax2.semilogx(analysis.frequency, np.absolute(analysis.out/current))
ax2.grid(True)
ax2.grid(True, which='minor')
ax2.set_xlabel("Frequency [Hz]")
ax2.set_ylabel('Rd [Ω]')

####################################################################################################

#r# We will now drive the diode with a pulse generator and perform a transient analysis.

#f# circuit_macros('diode-characteristic-curve-circuit-pulse.m4')

frequency = 1@u_MHz

circuit = Circuit('Diode')
circuit.include(spice_library['BAV21'])
# source = circuit.SinusoidalVoltageSource('input', 'in', circuit.gnd,
#                             dc_offset=dc_offset, offset=dc_offset,
#                             amplitude=ac_amplitude,
#                             frequency=frequency)
source = circuit.PulseVoltageSource('input', 'in', circuit.gnd,
                       initial_value=dc_offset-ac_amplitude, pulsed_value=dc_offset+ac_amplitude,
                       pulse_width=frequency.period/2, period=frequency.period)
circuit.R(1, 'in', 'out', 1@u_kΩ)
circuit.D('1', 'out', circuit.gnd, model='BAV21')

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=source.period/1e3, end_time=source.period*4)

# Fixme: axis, x scale
# plot(analysis['in'] - dc_offset + quiescent_points[0]['quiescent_voltage'])
# plot(analysis.out)
ax3.plot(analysis.out.abscissa*1e6, analysis.out)
ax3.legend(('Vin [V]', 'Vout [V]'), loc=(.8,.8))
ax3.grid()
ax3.set_xlabel('t [μs]')
ax3.set_ylabel('[V]')
# ax3.set_ylim(.5, 1 + ac_amplitude + .1)

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'diode-recovery-time.png')

#r# We notice the output of the circuit cannot follow the pulse generator.  It is due to the
#r# capacitive behaviour of a PN junction that cut off the highest frequencies of the pulse.  The
#r# plot of the dynamic resistance as a function of the frequency show a typical low pass filter
#r# behaviour where the impedance drop at high frequencies.



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2017 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import math
import unittest

import numpy as np
from numpy import testing as np_test

####################################################################################################

# Initialise logging
import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Probe.WaveForm import *
from PySpice.Unit import *

####################################################################################################

def print_rule():
    logger.info('\n' + '-'*100)

####################################################################################################

class TestUnits(unittest.TestCase):

    ##############################################

    @staticmethod
    def _test_unit_values(values, true_array):
        np_test.assert_almost_equal(values.as_ndarray(True), true_array)

    ##############################################

    # @unittest.skip('')
    def test(self):

        np_array1 = np.arange(10)
        array1 = u_mV(np_array1)
        np_raw_array1 = np_array1 / 1000

        np_array2 = np.arange(10, 20)
        array2 = u_mV(np_array2)
        np_raw_array2 = np_array2 / 1000

        print_rule()
        name = 'waveform1'
        title = 'A title'
        waveform1 = WaveForm(
            name, array1.prefixed_unit, array1.shape,
            title=title,
            abscissa=np_raw_array2,
        )
        waveform1[...] = array1
        logger.info(repr(waveform1))
        self._test_unit_values(waveform1, np_raw_array1)
        self.assertEqual(waveform1.prefixed_unit, array1.prefixed_unit)
        self.assertEqual(waveform1.name, name)
        self.assertEqual(waveform1.title, title)
        np_test.assert_array_equal(waveform1.abscissa, np_raw_array2)

        print_rule()
        waveform1 = WaveForm.from_unit_values(name, array1)
        logger.info(repr(waveform1))
        self._test_unit_values(waveform1, np_raw_array1)
        self.assertEqual(waveform1.prefixed_unit, array1.prefixed_unit)
        self.assertEqual(waveform1.name, name)

        print_rule()
        name = 'waveform2'
        waveform2 = WaveForm.from_unit_values(name, array2)
        logger.info(repr(waveform2))
        self._test_unit_values(waveform2, np_raw_array2)
        self.assertEqual(waveform2.prefixed_unit, array2.prefixed_unit)
        self.assertEqual(waveform2.name, name)

        print_rule()
        waveform3 = waveform2 - waveform1
        logger.info(repr(waveform3))
        self._test_unit_values(waveform3, np_raw_array2 - np_raw_array1)
        self.assertEqual(waveform3.prefixed_unit, array1.prefixed_unit)
        self.assertEqual(waveform3.name, '')

        print_rule()
        waveform4 = waveform3.convert(U_uV)
        logger.info('{} {}'.format(waveform4, type(waveform4)))

        print_rule()
        # Fixme:
        #     waveform5 = waveform1 >= 5
        #   File "PySpice/Probe/WaveForm.py", line 131, in __array_ufunc__
        #     result = super().__array_ufunc__(ufunc, method, *inputs, **kwargs)
        #   File "PySpice/Unit/Unit.py", line 1635, in __array_ufunc__
        #     raise ValueError
        ndarray5 = waveform1 >= 5@u_V
        np_test.assert_array_equal(ndarray5, np_raw_array1 >= 5)

        print_rule()
        waveform4 = waveform2 * waveform1
        logger.info(repr(waveform4))
        self.assertEqual(waveform4.name, '')
        self._test_unit_values(waveform4, np_raw_array2 * np_raw_array1)
        # Fixme: TypeError: unsupported operand type(s) for *: 'PrefixedUnit' and 'PrefixedUnit'
        # self.assertEqual(waveform4.prefixed_unit, array1.prefixed_unit * array2.prefixed_unit)
        mV2 = u_mV(1)*u_mV(1)
        self.assertEqual(waveform4.prefixed_unit.unit, mV2.unit)
        self.assertEqual(waveform4.prefixed_unit.power, mV2.power)

        print_rule()
        _ = waveform1 + waveform2
        waveform_mean = np.mean(_)
        logger.info(repr(waveform_mean))
        self._test_unit_values(waveform_mean, np.mean(np_raw_array1 + np_raw_array2))
        _ = u_mV(1)
        self.assertEqual(waveform_mean.unit, _.unit)
        self.assertEqual(waveform_mean.power, _.power)

####################################################################################################

if __name__ == '__main__':

    unittest.main()



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This modules implements the machinery to define element's parameters as descriptors.

"""

####################################################################################################

from ..Unit import Unit
from ..Tools.StringTools import str_spice

####################################################################################################

class ParameterDescriptor:

    """This base class implements a descriptor for element parameters.

    Public Attributes:

      :attr:`attribute_name`
        Name of the attribute in the element's class

      :attr:`default_value`
        The default value

    """

    ##############################################

    def __init__(self, default=None):

        self._default_value = default
        self._attribute_name = None

    ##############################################

    @property
    def default_value(self):
        return self._default_value

    @property
    def attribute_name(self):
        return self._attribute_name

    @attribute_name.setter
    def attribute_name(self, name):
        self._attribute_name = name

    ##############################################

    def __get__(self, instance, owner=None):

        try:
            return getattr(instance, '_' + self._attribute_name)
        except AttributeError:
            return self.default_value

    ##############################################

    def __set__(self, instance, value):
        setattr(instance, '_' + self._attribute_name, value)

    ##############################################

    def __repr__(self):
        return self.__class__.__name__

    ##############################################

    def validate(self, value):

        """Validate the parameter's value."""

        return value

    ##############################################

    def nonzero(self, instance):
        return self.__get__(instance) is not None

    ##############################################

    def to_str(self, instance):

        """Convert the parameter's value to SPICE syntax."""

        raise NotImplementedError

    ##############################################

    def __lt__(self, other):
        return self._attribute_name < other.attribute_name

####################################################################################################

class PositionalElementParameter(ParameterDescriptor):

    """This class implements a descriptor for positional element parameters.

    Public Attributes:

      :attr:`key_parameter`
        Flag to specify if the parameter is passed as key parameter in Python

      :attr:`position`
        Position of the parameter in the element definition

    """

    ##############################################

    def __init__(self, position, default=None, key_parameter=False):

        super().__init__(default)

        self._position = position
        self._key_parameter = key_parameter

    ##############################################

    @property
    def position(self):
        return self._position

    @property
    def key_parameter(self):
        return self._key_parameter

    ##############################################

    def to_str(self, instance):
        return str_spice(self.__get__(instance))

    ##############################################

    def __lt__(self, other):
        return self._position < other.position

####################################################################################################

class ElementNamePositionalParameter(PositionalElementParameter):

    """This class implements an element name positional parameter."""

    ##############################################

    def validate(self, value):
        return str(value)

####################################################################################################

class ExpressionPositionalParameter(PositionalElementParameter):

    """This class implements an expression positional parameter. """

    ##############################################

    def validate(self, value):
        return str(value)

####################################################################################################

class FloatPositionalParameter(PositionalElementParameter):

    """This class implements a float positional parameter."""

    ##############################################

    def __init__(self, position, unit=None, **kwargs):

        super().__init__(position, **kwargs)
        self._unit = unit

    ##############################################

    def validate(self, value):

        if isinstance(value, Unit):
            return value
        else:
            return Unit(value)

####################################################################################################

class InitialStatePositionalParameter(PositionalElementParameter):

    """This class implements an initial state (on, off) positional parameter."""

    ##############################################

    def validate(self, value):
        return bool(value) # Fixme: check KeyParameter

    ##############################################

    def to_str(self, instance):

        if self.__get__(instance):
            return 'on'
        else:
            return 'off'

####################################################################################################

class ModelPositionalParameter(PositionalElementParameter):

    """This class implements a model positional parameter. """

    ##############################################

    def validate(self, value):
        return str(value)

####################################################################################################

class FlagParameter(ParameterDescriptor):

    """This class implements a flag parameter.

    Public Attributes:

      :attr:`spice_name`
        Name of the parameter

    """

    ##############################################

    def __init__(self, spice_name, default=False):

        super().__init__(default)

        self.spice_name = spice_name

    ##############################################

    def nonzero(self, instance):
        return bool(self.__get__(instance))

    ##############################################

    def to_str(self, instance):

        if self.nonzero(instance):
            return 'off'
        else:
            return ''

####################################################################################################

class KeyValueParameter(ParameterDescriptor):

    """This class implements a key value pair parameter.

    Public Attributes:

      :attr:`spice_name`
        Name of the parameter

    """

    ##############################################

    def __init__(self, spice_name, default=None):

        super().__init__(default)

        self.spice_name = spice_name

    ##############################################

    def str_value(self, instance):
        return str_spice(self.__get__(instance))

    ##############################################

    def to_str(self, instance):

        if bool(self):
            return '{}={}'.format(self.spice_name, self.str_value(instance))
        else:
            return ''

####################################################################################################

class BoolKeyParameter(KeyValueParameter):

    """This class implements a boolean key parameter."""

    ##############################################

    def nonzero(self, instance):
        return bool(self.__get__(instance))

    ##############################################

    def str_value(self, instance):

        if self.nonzero(instance):
            return '1'
        else:
            return '0'

####################################################################################################

class ExpressionKeyParameter(KeyValueParameter):

    """This class implements an expression key parameter."""

    ##############################################

    def validate(self, value):
        return str(value)

####################################################################################################

class FloatKeyParameter(KeyValueParameter):

    """This class implements a float key parameter."""

    ##############################################

    def __init__(self, spice_name, unit=None, **kwargs):

        super().__init__(spice_name, **kwargs)
        self._unit = unit

    ##############################################

    def validate(self, value):
        return float(value)

####################################################################################################

class FloatPairKeyParameter(KeyValueParameter):

    """This class implements a float pair key parameter. """

    ##############################################

    def validate(self, pair):

        if len(pair) == 2:
            return (float(pair[0]), float(pair[1]))
        else:
            raise ValueError()

    ##############################################

    def str_value(self, instance):
        return ','.join([str(value) for value in self.__get__(instance)])

####################################################################################################

class FloatTripletKeyParameter(FloatPairKeyParameter):

    """This class implements a triplet key parameter."""

    ##############################################

    def validate(self, uplet):

        if len(uplet) == 3:
            return (float(uplet[0]), float(uplet[1]), float(uplet[2]))
        else:
            raise ValueError()

####################################################################################################

class IntKeyParameter(KeyValueParameter):

    """This class implements an integer key parameter."""

    ##############################################

    def validate(self, value):
        return int(value)



####################################################################################################

from PySpice.Spice.Netlist import SubCircuitFactory
from PySpice.Unit import *

####################################################################################################

class BasicOperationalAmplifier(SubCircuitFactory):

    NAME = 'BasicOperationalAmplifier'
    NODES = ('non_inverting_input', 'inverting_input', 'output')

    ##############################################

    def __init__(self):

        super().__init__()

        # Input impedance
        self.R('input', 'non_inverting_input', 'inverting_input', 10@u_MΩ)

        # dc gain=100k and pole1=100hz
        # unity gain = dcgain x pole1 = 10MHZ
        self.VCVS('gain', 1, self.gnd, 'non_inverting_input', 'inverting_input', voltage_gain=kilo(100))
        self.R('P1', 1, 2, 1@u_kΩ)
        self.C('P1', 2, self.gnd, 1.5915@u_uF)

        # Output buffer and resistance
        self.VCVS('buffer', 3, self.gnd, 2, self.gnd, 1)
        self.R('out', 3, 'output', 10@u_Ω)

####################################################################################################

class BasicComparator(SubCircuitFactory):

    NAME = 'BasicComparator'
    NODES = ('non_inverting_input', 'inverting_input',
             'voltage_plus', 'voltage_minus',
             'output')

    ##############################################

    def __init__(self,):

        super().__init__()

        # Fixme: ngspice is buggy with such subcircuit

        # Fixme: how to pass voltage_plus, voltage_minus ?
        # output_voltage_minus, output_voltage_plus = 0, 15

        # to plug the voltage source
        self.R(1, 'voltage_plus', 'voltage_minus', 1@u_MΩ)
        self.NonLinearVoltageSource(1, 'output', 'voltage_minus',
                                    expression='V(non_inverting_input, inverting_input)',
                                    # table=((-micro(1), output_voltage_minus),
                                    #       (micro(1), output_voltage_plus))
                                    table=(('-1uV', '0V'), ('1uV', '15V'))
                                )



####################################################################################################

#r#
#r# ========================================
#r#  Spice Netlist Parser Bootstrap Example
#r# ========================================
#r#
#r# This example shows a bootstrap of a netlist, i.e. we parse the netlist generated by PySpice
#r# and we regenerate it.
#r#

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Doc.ExampleTools import find_libraries
from PySpice.Spice.Library import SpiceLibrary
from PySpice.Spice.Netlist import Circuit
from PySpice.Spice.Parser import SpiceParser
from PySpice.Unit import *

####################################################################################################

libraries_path = find_libraries()
spice_library = SpiceLibrary(libraries_path)

####################################################################################################

circuit = Circuit('STM AN1476: Low-Cost Power Supply For Home Appliances')

circuit.include(spice_library['1N4148'])
# 1N5919B: 5.6 V, 3.0 W Zener Diode Voltage Regulator
circuit.include(spice_library['d1n5919brl'])

ac_line = circuit.AcLine('input', 'out', 'in', rms_voltage=230, frequency=50)
circuit.R('load', 'out', circuit.gnd, kilo(1))
circuit.C('load', 'out', circuit.gnd, micro(220))
circuit.X('D1', '1N4148', circuit.gnd, 1)
circuit.D(1, circuit.gnd, 1, model='DIODE1', off=True)
circuit.X('Dz1', 'd1n5919brl', 1, 'out')
circuit.C('ac', 1, 2, nano(470))
circuit.R('ac', 2, 'in', 470) # Fixme: , m=1, temperature='{25}'

source = str(circuit)
print(source)

####################################################################################################

parser = SpiceParser(source=source)
bootstrap_circuit = parser.build_circuit()

bootstrap_source = str(bootstrap_circuit)
print(bootstrap_source)

assert(source == bootstrap_source)



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

import logging

from . import BasicElement
from . import HighLevelElement
from .Netlist import Netlist, ElementParameterMetaClass

####################################################################################################

_module_logger = logging.getLogger(__name__)

####################################################################################################

def _get_elements(module):
    element_classes = []
    for item  in module.__dict__.values():
        if (type(item) is ElementParameterMetaClass
            and item.PREFIX is not None
           ):
            element_classes.append(item)
    return element_classes

####################################################################################################
#
# Add a method to create elements to the Netlist class
#

spice_elements = _get_elements(BasicElement)
high_level_elements = _get_elements(HighLevelElement)

for element_class in spice_elements + high_level_elements:

    def _make_function(element_class):
        def function(self, *args, **kwargs):
            return element_class(self, *args, **kwargs)
        # Preserve docstrings for element shortcuts
        function.__doc__ = element_class.__doc__
        return function

    func = _make_function(element_class)

    def _set(name):
        # _module_logger.debug("Add device shortcut {} for class {}".format(name, element_class))
        setattr(Netlist, name, func)

    _set(element_class.__name__)

    if element_class in spice_elements:
        if hasattr(element_class, 'ALIAS'):
            _set(element_class.ALIAS)
        if hasattr(element_class, 'LONG_ALIAS'):
            _set(element_class.LONG_ALIAS)





#r# This example shows the simulation of a transmission line.

####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Probe.Plot import plot
from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

####################################################################################################

#r# We will drive the transmission line with a pulse source and use a standard 50 Ω load.

circuit = Circuit('Transmission Line')
circuit.PulseVoltageSource('pulse', 'input', circuit.gnd, 0@u_V, 1@u_V, 1@u_ns, 1@u_us)
circuit.LosslessTransmissionLine('delay', 'output', circuit.gnd, 'input', circuit.gnd,
                                 impedance=50, time_delay=40e-9)
circuit.R('load', 'output', circuit.gnd, 50@u_Ω)

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
analysis = simulator.transient(step_time=1e-11, end_time=100e-9)

####################################################################################################

figure, ax = plt.subplots(figsize=(20, 6))
ax.plot(analysis['input'])
ax.plot(analysis['output'])
ax.set_xlabel('Time [s]')
ax.set_ylabel('Voltage (V)')
ax.grid()
ax.legend(['input', 'output'], loc='upper right')

plt.show()

#f# save_figure('figure', 'time-delay.png')



#skip#

####################################################################################################

import time

import numpy as np

from cffi import FFI

####################################################################################################

ffi = FFI()

####################################################################################################

ffi.cdef("""
typedef struct ngcomplex
{
  double cx_real;
  double cx_imag;
} ngcomplex_t;

typedef struct vector_info
{
  char *v_name;
  int v_type;
  short v_flags;
  double *v_realdata;
  ngcomplex_t *v_compdata;
  int v_length;
} vector_info, *pvector_info;

typedef struct vecvalues
{
  char *name;
  double creal;
  double cimag;
  bool is_scale;
  bool is_complex;
} vecvalues, *pvecvalues;

typedef struct vecvaluesall
{
  int veccount;
  int vecindex;
  pvecvalues *vecsa;
} vecvaluesall, *pvecvaluesall;

typedef struct vecinfo
{
  int number;
  char *vecname;
  bool is_real;
  void *pdvec;
  void *pdvecscale;
} vecinfo, *pvecinfo;

typedef struct vecinfoall
{
  char *name;
  char *title;
  char *date;
  char *type;
  int veccount;
  pvecinfo *vecs;
} vecinfoall, *pvecinfoall;

typedef int (SendChar) (char *, int, void *);
typedef int (SendStat) (char *, int, void *);
typedef int (ControlledExit) (int, bool, bool, int, void *);
typedef int (SendData) (pvecvaluesall, int, int, void *);
typedef int (SendInitData) (pvecinfoall, int, void *);
typedef int (BGThreadRunning) (bool, int, void *);
typedef int (GetVSRCData) (double *, double, char *, int, void *);
typedef int (GetISRCData) (double *, double, char *, int, void *);
typedef int (GetSyncData) (double, double *, double, int, int, int, void *);

int ngSpice_Init (SendChar *, SendStat *, ControlledExit *, SendData *, SendInitData *, BGThreadRunning *, void *);
int ngSpice_Init_Sync (GetVSRCData *, GetISRCData *, GetSyncData *, int *, void *);

int ngSpice_Command (char *);
pvector_info ngGet_Vec_Info (char *);
int ngSpice_Circ (char **);
char *ngSpice_CurPlot (void);
char **ngSpice_AllPlots (void);
char **ngSpice_AllVecs (char *);
bool ngSpice_running (void);
bool ngSpice_SetBkpt (double);
""")

####################################################################################################

ngspice_shared_path = '/usr/local/stow/ngspice-26/lib/libngspice.so'
ngspice_shared = ffi.dlopen(ngspice_shared_path)

####################################################################################################

@ffi.callback("int (char *, int, void *)")
def send_char_callback(message, ngspice_id, user_data):
    print('>>> send_char_callback', ffi.string(message), ngspice_id)
    return 0
#send_char_callback = ffi.NULL

@ffi.callback("int (char *, int, void *)")
def send_stat_callback(message, ngspice_id, user_data):
    print('>>> send_stat_callback', ffi.string(message), ngspice_id)
    return 0
#send_stat_callback = ffi.NULL

@ffi.callback("int (int, bool, bool, int, void *)")
def exit_callback(exit_status, immediate_unloding, quit_exit, ngspice_id, user_data):
    print("exit_callback", exit_status, immediate_unloding, quit_exit, ngspice_id)
    return exit_status

@ffi.callback("int (pvecvaluesall, int, int, void *)")
def send_data_callback(data, number_of_vectors, ngspice_id, user_data):
    print(">>> send_data [{}]".format(data.vecindex))

    for i in range(int(number_of_vectors)):
        actual_vector_value = data.vecsa[i]
        print("    Vector: {} {} +i {}".format(ffi.string(actual_vector_value.name),
                                               actual_vector_value.creal,
                                               actual_vector_value.cimag))

    return 0
#send_data_callback = ffi.NULL

@ffi.callback("int (pvecinfoall, int, void *)")
def send_init_data_callback(data,  ngspice_id, user_data):
    print(">>> send_init_data")
    number_of_vectors = data.veccount
    for i in range(number_of_vectors):
        print("  Vector:", ffi.string(data.vecs[i].vecname))

    return 0
#send_init_data_callback = ffi.NULL

@ffi.callback("int (double *, double, char *, int, void *)")
def get_vsrc_data(voltage, time, node, ngspice_id, user_data):
    print(">>> get_vsrc_data @{} node {}".format(time, ffi.string(node)))
    voltage[0] = 1.
    return 0

####################################################################################################

rc = ngspice_shared.ngSpice_Init(send_char_callback,
                                 send_stat_callback,
                                 exit_callback,
                                 send_data_callback,
                                 send_init_data_callback,
                                 ffi.NULL,
                                 ffi.NULL)
print('ngSpice_Init returned', rc)

ngspice_id = ffi.new('int *', 0)
rc = ngspice_shared.ngSpice_Init_Sync(get_vsrc_data, ffi.NULL, ffi.NULL, ngspice_id, ffi.NULL)
print("ngSpice_Init_Sync returned: ", rc)

time_step = '100u'

circuit = [
    ".title rc circuit",
    # "V1 1 0 1",
    "V1 1 0 dc 0 external",
    "R1 1 2 1",
    "C1 2 0 1 ic=0",
    ".tran {} 3 uic".format(time_step),
    ".end",
]

# for line in circuit:
#     rc = ngspice_shared.ngSpice_Command(('circbyline ' + line).encode('utf8'))
# print('ngSpice_Command', rc)

circuit_lines_keepalive = [ffi.new("char[]", line.encode('utf8')) for line in circuit] + [ffi.NULL]
circuit_array = ffi.new("char *[]", circuit_lines_keepalive)
rc = ngspice_shared.ngSpice_Circ(circuit_array)
print('ngSpice_Circ', rc)

rc = ngspice_shared.ngSpice_Command('bg_run'.encode('utf8'))
print('ngSpice_Command returned', rc)

time.sleep(.1) # required before to test if the simulation is running
while (ngspice_shared.ngSpice_running()):
    time.sleep(.1)
print("Simulation is done")

all_plots = ngspice_shared.ngSpice_AllPlots()
i = 0
while (True):
    if all_plots[i] == ffi.NULL:
        break
    else:
        print(ffi.string(all_plots[i]))
    i += 1

plot_name = 'tran1'
all_vectors = ngspice_shared.ngSpice_AllVecs(plot_name.encode('utf8'))
i = 0
while (True):
    if all_vectors[i] == ffi.NULL:
        break
    else:
        vector_name = ffi.string(all_vectors[i])
        name = '.'.join((plot_name, vector_name.decode('utf8')))
        vector_info = ngspice_shared.ngGet_Vec_Info(name.encode('utf8'))
        length = vector_info.v_length
        print("vector[{}] {} type {} flags {} length {}".format(i,
                                                                vector_name,
                                                                vector_info.v_type,
                                                                vector_info.v_flags,
                                                                length))
        if vector_info.v_compdata == ffi.NULL:
            print("  real data")
            # for k in range(length):
            #     print("  [{}] {}".format(k, vector_info.v_realdata[k]))
            real_array = np.frombuffer(ffi.buffer(vector_info.v_realdata, length*8), dtype=np.float64)
            print(real_array)
        else:
            print("  complex data")
            for k in range(length):
                value = vector_info.v_compdata[k]
                print("  [{}] {} + i {}".format(k, value.cx_real, value.cx_imag))
    i += 1



####################################################################################################
#
# PySpice - A Spice package for Python
# Copyright (C) 2020 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

from pathlib import Path
import shutil

from invoke import task

####################################################################################################

# https://github.com/conda-forge/pyspice-feedstock

# enable anaconda environment
#   source .../miniconda3/bin/activate

@task()
def conda_build(ctx):
    with ctx.cd('anaconda'):
        ctx.run('conda build .')
        # ctx.run('conda build purge')

@task()
def conda_login(ctx):
    path = Path(__file__).parent.joinpath('anaconda-login.txt')
    with open(path) as fh:
        username = fh.readline().strip()
        password = fh.readline().strip()
    print('"{}" "{}"'.format(username, password))
    ctx.run('anaconda login --username="{}" --password="{}"'.format(username, password))

@task(conda_build, conda_login)
def conda_upload(ctx):
    result = ctx.run('conda build . --output')
    path = str(result.stdout).strip()
    ctx.run('anaconda upload {}'.format(path))



####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

"""This module provides helpers to plot Bode diagrams using Matplolib.

Frequency is in Hz, gain in dB, phase in radians between -π and π.
"""

####################################################################################################

import math

from matplotlib import pyplot

####################################################################################################

def bode_diagram_gain(axe, frequency, gain, **kwargs):

    axe.semilogx(frequency, gain, base=10, **kwargs)
    axe.grid(True)
    axe.grid(True, which='minor')
    axe.set_xlabel("Frequency [Hz]")
    axe.set_ylabel("Gain [dB]")

####################################################################################################

def bode_diagram_phase(axe, frequency, phase, **kwargs):

    axe.semilogx(frequency, phase, base=10, **kwargs)
    axe.set_ylim(-math.pi, math.pi)
    axe.grid(True)
    axe.grid(True, which='minor')
    axe.set_xlabel("Frequency [Hz]")
    axe.set_ylabel("Phase [rads]")
    # axe.set_yticks # Fixme:
    pyplot.yticks((-math.pi, -math.pi/2,0, math.pi/2, math.pi),
                  (r"$-\pi$", r"$-\frac{\pi}{2}$", "0", r"$\frac{\pi}{2}$", r"$\pi$"))

####################################################################################################

def bode_diagram(axes, frequency, gain, phase, **kwargs):
    bode_diagram_gain(axes[0], frequency, gain, **kwargs)
    bode_diagram_phase(axes[1], frequency, phase, **kwargs)



#skip#

####################################################################################################

import matplotlib.pyplot as plt

####################################################################################################

import PySpice.Logging.Logging as Logging
logger = Logging.setup_logging()

####################################################################################################

from PySpice.Spice.Netlist import Circuit
from PySpice.Unit import *

# from OperationalAmplifier import basic_comparator

####################################################################################################

circuit = Circuit('Astable Multivibrator')

source = circuit.V('cc', 'vcc', circuit.gnd, 15@u_V)
# Time constant
circuit.R(1, 'output', 'comparator', 1@u_kΩ)
circuit.C(1, 'comparator', circuit.gnd, 100@u_nF)
# Reference
circuit.R(2, 'output', 'reference', 100@u_kΩ)
circuit.R(3, 'vcc', 'reference', 100@u_kΩ)
circuit.R(4, 'reference', circuit.gnd, 100@u_kΩ)
# Comparator
# Fixme: ngspice is buggy with such subcircuit
# circuit.subcircuit(basic_comparator)
# circuit.X('comparator', 'BasicComparator', 'reference', 'comparator', 'vcc', circuit.gnd, 'output')
circuit.NonLinearVoltageSource(1, 'output', circuit.gnd,
                               expression='V(reference, comparator)',
                               table=((-micro(1), 0),
                                      (micro(1), source.dc_value))
                               )

simulator = circuit.simulator(temperature=25, nominal_temperature=25)
# simulator.initial_condition(comparator=0)  # Fixme: simulator.nodes.comparator == 0
simulator.node_set(comparator=0)  # Fixme: simulator.nodes.comparator == 0
analysis = simulator.transient(step_time=1@u_us, end_time=500@u_us)
# Fixme: Xyce fails with Time step too small

figure, ax = plt.subplots(figsize=(20, 10))
ax.grid()
ax.plot(analysis.reference)
ax.plot(analysis.comparator)
ax.plot(analysis.output)

plt.tight_layout()
plt.show()

#f# save_figure('figure', 'astable.png')



#skip#

####################################################################################################

from PySpice.Spice.Netlist import SubCircuit
from PySpice.Unit import *

####################################################################################################

class BasicOperationalAmplifier(SubCircuit): # SubCircuitFactory

    #
    # __init__(self, name, subcircuit_name, *nodes)
    #
    # name = class name
    # node = NODES = interface
    #

    NODES = ('non_inverting_input', 'inverting_input', 'output')

    ##############################################

    def __init__(self):

        # comment: we could pass parameters using ctor

        # Input impedance
        # comment: 'R'+'2' but for other devices ? name/attribute versus spice name
        self.R('input', 'non_inverting_input', 'inverting_input', 10@u_MΩ)

        # dc gain=100k and pole1=100hz
        # unity gain = dcgain x pole1 = 10MHZ
        # Fixme: gain=...
        self.VCVS('gain', 'non_inverting_input', 'inverting_input', 1, self.gnd, kilo(100))
        self.R('P1', 1, 2, 1@u_kΩ)
        self.C('P1', 2, self.gnd, 1.5915@u_uF)

        # Output buffer and resistance
        self.VCVS('buffer', 2, self.gnd, 3, self.gnd, 1)
        self.R('out', 3, 'output', 10@u_Ω)

####################################################################################################

class BasicOperationalAmplifier(SubCircuit): # SubCircuitFactory

    NODES = ('non_inverting_input', 'inverting_input', 'output')

    # Comment: R doesn't know its name, R prefix is redundant
    Rinput = R('non_inverting_input', 'inverting_input', 10@u_MΩ)

    gain = VCVS('non_inverting_input', 'inverting_input', 1, self.gnd, kilo(100))
    RP1 = R(1, 2, 1@u_kΩ)
    CP1 = C(2, self.gnd, 1.591@u_uF)

    # Comment: buffer is a Python name
    buffer = VCVS(2, self.gnd, 3, self.gnd, 1)
    Rout = R(3, 'output', 10@u_Ω)



# -*- coding: utf-8 -*-

####################################################################################################
#
# PySpice - A Spice Package for Python
# Copyright (C) 2014 Fabrice Salvaire
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
#
####################################################################################################

####################################################################################################

class Copper:

    atomic_number = 29

    atomic_mass = 63.546 * 1e-3 # kg
    density = 8.96 * 1e3 # kg·m−3
    thermal_conductivity = 401 # W·m−1·K−1
    electrical_resistivity = 16.78 * 1e-9 # Ω·m @20 °C
    electron_mobility = - 4.6 * 1e3 # m2·V−1·s−1

    ##############################################

    def electrical_resistance_for_conductor(self, degree):
        """ Used to compute conductor resistance. """
        rho0 = 16e-3 # Ω·m·mm−2
        return rho0 * (1 + .00393 * degree)
